# -*- coding: utf-8 -*-
# Copyright © 2014-2017 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Define API endpoints for managing and searching buildroot overrides."""

import math
from datetime import datetime

from cornice import Service
from cornice.validators import colander_body_validator, colander_querystring_validator
from pyramid.exceptions import HTTPNotFound

from sqlalchemy import func, distinct
from sqlalchemy.sql import or_

from bodhi.server import log, security
from bodhi.server.models import Build, BuildrootOverride, Package, Release, User
import bodhi.server.schemas
import bodhi.server.services.errors
from bodhi.server.validators import (
    validate_override_builds,
    validate_expiration_date,
    validate_packages,
    validate_releases,
    validate_username,
)


override = Service(name='override', path='/overrides/{nvr}',
                   description='Buildroot Overrides',
                   cors_origins=bodhi.server.security.cors_origins_ro)

overrides = Service(name='overrides', path='/overrides/',
                    description='Buildroot Overrides',
                    factory=security.PackagerACLFactory,
                    # Note, this 'rw' is not a typo.  the @comments service has
                    # a ``post`` section at the bottom.
                    cors_origins=bodhi.server.security.cors_origins_rw)

overrides_rss = Service(name='overrides_rss', path='/rss/overrides/',
                        description='Buildroot Overrides RSS Feed',
                        cors_origins=bodhi.server.security.cors_origins_ro)


@override.get(accept=("application/json", "text/json"), renderer="json",
              error_handler=bodhi.server.services.errors.json_handler)
@override.get(accept=("application/javascript"), renderer="jsonp",
              error_handler=bodhi.server.services.errors.jsonp_handler)
@override.get(accept=("text/html"), renderer="override.html",
              error_handler=bodhi.server.services.errors.html_handler)
def get_override(request):
    """
    Return a dictionary with key "override" indexing the override that matches the given nvr.

    Args:
        request (pyramid.request): The current request, which should have a query parameter "nvr",
            providing the nvr of the requested override.
    Returns:
        dict: A dictionary with key "override" that indexes the override matching the given nvr.
    """
    nvr = request.matchdict.get('nvr')

    build = Build.get(nvr)

    if not build:
        request.errors.add('url', 'nvr', 'No such build')
        request.errors.status = HTTPNotFound.code
        return

    if not build.override:
        request.errors.add('url', 'nvr',
                           'No buildroot override for this build')
        request.errors.status = HTTPNotFound.code
        return

    return dict(override=build.override)


validators = (
    colander_querystring_validator,
    validate_packages,
    validate_releases,
    validate_username,
)


@overrides_rss.get(schema=bodhi.server.schemas.ListOverrideSchema, renderer='rss',
                   error_handler=bodhi.server.services.errors.html_handler,
                   validators=validators)
@overrides.get(schema=bodhi.server.schemas.ListOverrideSchema, renderer='rss',
               accept=('application/atom+xml',),
               error_handler=bodhi.server.services.errors.html_handler,
               validators=validators)
@overrides.get(schema=bodhi.server.schemas.ListOverrideSchema,
               accept=("application/json", "text/json"), renderer="json",
               error_handler=bodhi.server.services.errors.json_handler,
               validators=validators)
@overrides.get(schema=bodhi.server.schemas.ListOverrideSchema,
               accept=("application/javascript"), renderer="jsonp",
               error_handler=bodhi.server.services.errors.jsonp_handler,
               validators=validators)
@overrides.get(schema=bodhi.server.schemas.ListOverrideSchema,
               accept=('text/html'), renderer='overrides.html',
               error_handler=bodhi.server.services.errors.html_handler,
               validators=validators)
def query_overrides(request):
    """
    Search for overrides by various criteria.

    The following optional parameters may be used when searching for overrides:
        builds (list): A list of NVRs to search overrides by.
        expired (bool): If True, limit search to expired overrides. If False, limit search to active
            overrides.
        like (basestring): Perform an SQL "like" query against build NVRs with the given string.
        packages (list): A list of package names to search overrides by.
        releases (list): A list of release names to limit the overrides search by.
        search (basestring): Perform an SQL "ilike" query against build NVRs with the given string.
        submitter (basestring): Search for overrides submitted by the given username.

    Returns:
        dict: A dictionary with the following keys:
            overrides: An iterable containing the matched overrides.
            page: The current page number in the results.
            pages: The number of pages of results that match the query.
            rows_per_page: The number of rows on the page.
            total: The total number of overrides that match the criteria.
            chrome: The caller supplied chrome.
            display_user: The current username.
    """
    db = request.db
    data = request.validated
    query = db.query(BuildrootOverride)

    expired = data.get('expired')
    if expired is not None:
        if expired:
            query = query.filter(BuildrootOverride.expired_date.isnot(None))
        else:
            query = query.filter(BuildrootOverride.expired_date.is_(None))

    builds = data.get('builds')
    if builds is not None:
        query = query.join(BuildrootOverride.build)
        query = query.filter(or_(*[Build.nvr == bld for bld in builds]))

    packages = data.get('packages')
    if packages is not None:
        query = query.join(BuildrootOverride.build).join(Build.package)
        query = query.filter(or_(*[Package.name == pkg.name for pkg in packages]))

    releases = data.get('releases')
    if releases is not None:
        query = query.join(BuildrootOverride.build).join(Build.release)
        query = query.filter(or_(*[Release.name == r.name for r in releases]))

    like = data.get('like')
    if like is not None:
        query = query.join(BuildrootOverride.build)
        query = query.filter(or_(*[
            Build.nvr.like('%%%s%%' % like)
        ]))

    search = data.get('search')
    if search is not None:
        query = query.join(BuildrootOverride.build)
        query = query.filter(Build.nvr.ilike('%%%s%%' % search))

    submitter = data.get('user')
    if submitter is not None:
        query = query.filter(or_(*[BuildrootOverride.submitter == s for s in submitter]))

    query = query.order_by(BuildrootOverride.submission_date.desc())

    # We can't use ``query.count()`` here because it is naive with respect to
    # all the joins that we're doing above.
    count_query = query.with_labels().statement\
        .with_only_columns([func.count(distinct(BuildrootOverride.id))])\
        .order_by(None)
    total = db.execute(count_query).scalar()

    page = data.get('page')
    rows_per_page = data.get('rows_per_page')
    pages = int(math.ceil(total / float(rows_per_page)))
    query = query.offset(rows_per_page * (page - 1)).limit(rows_per_page)

    return dict(
        overrides=query.all(),
        page=page,
        pages=pages,
        rows_per_page=rows_per_page,
        total=total,
        chrome=data.get('chrome'),
        display_user=data.get('display_user'),
    )


@overrides.post(schema=bodhi.server.schemas.SaveOverrideSchema,
                permission='edit',
                accept=("application/json", "text/json"), renderer='json',
                error_handler=bodhi.server.services.errors.json_handler,
                validators=(
                    colander_body_validator,
                    validate_override_builds,
                    validate_expiration_date,
                ))
@overrides.post(schema=bodhi.server.schemas.SaveOverrideSchema,
                permission='edit',
                accept=("application/javascript"), renderer="jsonp",
                error_handler=bodhi.server.services.errors.jsonp_handler,
                validators=(
                    colander_body_validator,
                    validate_override_builds,
                    validate_expiration_date,
                ))
def save_override(request):
    """
    Create or edit a buildroot override.

    This entails either creating a new buildroot override, or editing an
    existing one. To edit an existing buildroot override, the buildroot
    override's original id needs to be specified in the ``edited`` parameter.

    Args:
        request (pyramid.request): The current web request.
    Returns:
        dict: The new or edited override.
    """
    data = request.validated

    edited = data.pop("edited")

    caveats = []
    try:
        submitter = User.get(request.user.name)
        if edited is None:
            builds = data['builds']
            overrides = []
            if len(builds) > 1:
                caveats.append({
                    'name': 'nvrs',
                    'description': 'Your override submission was '
                    'split into %i.' % len(builds)
                })
            for build in builds:
                log.info("Creating a new buildroot override: %s" % build.nvr)
                existing_override = BuildrootOverride.get(build.id)
                if existing_override:
                    if not existing_override.expired_date:
                        data['expiration_date'] = max(existing_override.expiration_date,
                                                      data['expiration_date'])

                    data['notes'] = """{0}

_@{1} ({2})_

_____________
{3}""".format(existing_override.notes, existing_override.submitter.name,
                        existing_override.submission_date.strftime("%b %d, %Y"), data['notes'])

                    overrides.append(BuildrootOverride.edit(
                        request,
                        edited=build,
                        submitter=submitter,
                        submission_date=datetime.now(),
                        notes=data['notes'],
                        expiration_date=data['expiration_date'],
                        expired=None,
                    ))
                else:
                    overrides.append(BuildrootOverride.new(
                        request,
                        build=build,
                        submitter=submitter,
                        notes=data['notes'],
                        expiration_date=data['expiration_date'],
                    ))

            if len(builds) > 1:
                result = dict(overrides=overrides)
            else:
                result = overrides[0]
        else:
            log.info("Editing buildroot override: %s" % edited)

            edited = Build.get(edited)

            if edited is None:
                request.errors.add('body', 'edited', 'No such build')
                return

            result = BuildrootOverride.edit(
                request, edited=edited, submitter=submitter,
                notes=data["notes"], expired=data["expired"],
                expiration_date=data["expiration_date"])

            if not result:
                # Some error inside .edit(...)
                return

    except Exception as e:
        log.exception(e)
        request.errors.add('body', 'override',
                           'Unable to save buildroot override: %s' % e)
        return

    if not isinstance(result, dict):
        result = result.__json__()

    result['caveats'] = caveats

    return result



# Copyright (c) 2018 Sebastian Wojciechowski
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""This script will run a handy python shell initialized with Bodhi models."""

import sys
from subprocess import call
import logging

import click

from bodhi.server.config import get_configfile


logger = logging.getLogger('bodhi-shell')
logging.basicConfig(level=logging.INFO)


@click.command()
@click.version_option(message='%(version)s')
def get_bodhi_shell():
    """Run python shell initialized with Bodhi models."""
    configfile = get_configfile()
    if configfile is not None:
        call(['pshell-3', configfile])
    else:
        click.echo("Config file not found!")
        sys.exit(1)



# -*- coding: utf-8 -*-
# Copyright © 2014-2017 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Define service endpoint for retrieving Builds."""
import math

from cornice import Service
from cornice.validators import colander_querystring_validator
from pyramid.exceptions import HTTPNotFound
from sqlalchemy import func, distinct
from sqlalchemy.sql import or_

from bodhi.server.models import Update, Build, Package, Release
from bodhi.server.validators import (validate_updates,
                                     validate_packages, validate_releases)
import bodhi.server.schemas
import bodhi.server.security
import bodhi.server.services.errors


build = Service(name='build', path='/builds/{nvr}', description='Koji builds',
                cors_origins=bodhi.server.security.cors_origins_ro)
builds = Service(name='builds', path='/builds/',
                 description='Koji builds',
                 cors_origins=bodhi.server.security.cors_origins_ro)


@build.get(renderer='json',
           error_handler=bodhi.server.services.errors.json_handler)
def get_build(request):
    """
    Retrieve a Build by name-version-release, specified via an "nvr" query string parameter.

    Args:
        request (pyramid.request): The current web request.
    Returns:
        bodhi.server.models.Build or None: The Build matching the search, or None if there is no
            Build with the given NVR.
    """
    nvr = request.matchdict.get('nvr')
    build = Build.get(nvr)
    if not build:
        request.errors.add('body', 'nvr', 'No such build')
        request.errors.status = HTTPNotFound.code
        return
    return build


@builds.get(schema=bodhi.server.schemas.ListBuildSchema, renderer='json',
            error_handler=bodhi.server.services.errors.json_handler,
            validators=(colander_querystring_validator, validate_releases, validate_updates,
                        validate_packages))
def query_builds(request):
    """
    Search for Builds by given criteria.

    The following criteria can be provided via query string parameters to search for Builds:
        nvr: The dash-separated name-version-release of a Build.
        updates: A space or comma separated list of updates to limit the search by.
        packages: A space or comma separated list of packages to search for builds by.
        releases: A space or comma separated list of release ids to limit builds by.
        page: Which page of search results are desired.
        rows_per_pags: How many results per page are desired.

    Args:
        request (pyramid.request): The current request, containing the search criteria documented
            above.
    Returns:
        dict: A dictionary with the following key value mappings:
            builds: An iterable of builds that match the search criteria.
            page: The current page.
            pages: The total number of pages.
            rows_per_page: The number of rows per page.
            total: The number of builds that match the search criteria.
    """
    db = request.db
    data = request.validated
    query = db.query(Build)

    nvr = data.get('nvr')
    if nvr is not None:
        query = query.filter(Build.nvr == nvr)

    updates = data.get('updates')
    if updates is not None:
        query = query.join(Build.update)
        args = \
            [Update.title == update.title for update in updates] +\
            [Update.alias == update.alias for update in updates]
        query = query.filter(or_(*args))

    packages = data.get('packages')
    if packages is not None:
        query = query.join(Build.package)
        query = query.filter(or_(*[Package.id == p.id for p in packages]))

    releases = data.get('releases')
    if releases is not None:
        query = query.join(Build.release)
        query = query.filter(or_(*[Release.id == r.id for r in releases]))

    # We can't use ``query.count()`` here because it is naive with respect to
    # all the joins that we're doing above.
    count_query = query.with_labels().statement\
        .with_only_columns([func.count(distinct(Build.nvr))])\
        .order_by(None)
    total = db.execute(count_query).scalar()

    page = data.get('page')
    rows_per_page = data.get('rows_per_page')
    pages = int(math.ceil(total / float(rows_per_page)))
    query = query.offset(rows_per_page * (page - 1)).limit(rows_per_page)

    return dict(
        builds=query.all(),
        page=page,
        pages=pages,
        rows_per_page=rows_per_page,
        total=total,
    )



# -*- coding: utf-8 -*-
# Copyright 2015-2018 Red Hat Inc., and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
The "updates handler".

This module is responsible for doing value-added work "offline" that used to be
done when updates were submitted.  Specifically, when someone submits an update
we used to:

- Update any bugs in bugzilla associated with the update.
- Check for test cases in the wiki.

Those things could sometimes take a *very* long time, especially if there were
lots of builds and lots of bugs in the update.

Now, update-submission breezes by those steps and simply tells the user "OK".
A fedmsg message gets published when their update goes through, and *that*
message gets received here and triggers us to do all that network-laden heavy
lifting.
"""

import logging
import pprint
import time

import fedmsg.consumers

from bodhi.server import initialize_db, util, bugs as bug_module
from bodhi.server.config import config
from bodhi.server.exceptions import BodhiException
from bodhi.server.models import Bug, Update, UpdateType


log = logging.getLogger('bodhi')


class UpdatesHandler(fedmsg.consumers.FedmsgConsumer):
    """
    Perform background tasks when updates are created or edited.

    This fedmsg listener waits for messages from the frontend about new or edited updates, and
    performs background tasks such as modifying Bugzilla issues (and loading information from
    Bugzilla so we can display it to the user) and looking up wiki test cases.

    Attributes:
        db_factory (bodhi.server.util.TransactionalSessionMaker): A context manager that yields a
            database session.
        handle_bugs (bool): If True, interact with Bugzilla. Else do not.
        topic (list): A list of strings that indicate which fedmsg topics this consumer listens to.
    """

    config_key = 'updates_handler'

    def __init__(self, hub, *args, **kwargs):
        """
        Initialize the UpdatesHandler, subscribing it to the appropriate topics.

        Args:
            hub (moksha.hub.hub.CentralMokshaHub): The hub this handler is consuming messages from.
                It is used to look up the hub config.
        """
        initialize_db(config)
        self.db_factory = util.transactional_session_maker()

        prefix = hub.config.get('topic_prefix')
        env = hub.config.get('environment')
        self.topic = [
            prefix + '.' + env + '.bodhi.update.request.testing',
            prefix + '.' + env + '.bodhi.update.edit',
        ]

        self.handle_bugs = bool(config.get('bodhi_email'))
        if not self.handle_bugs:
            log.warning("No bodhi_email defined; not fetching bug details")
        else:
            bug_module.set_bugtracker()

        super(UpdatesHandler, self).__init__(hub, *args, **kwargs)
        log.info('Bodhi updates handler listening on:\n'
                 '%s' % pprint.pformat(self.topic))

    def consume(self, message):
        """
        Process the given message, updating relevant bugs and test cases.

        Args:
            message (munch.Munch): A fedmsg about a new or edited update.
        """
        msg = message['body']['msg']
        topic = message['topic']
        alias = msg['update'].get('alias')

        log.info("Updates Handler handling  %s, %s" % (alias, topic))

        # Go to sleep for a second to try and avoid a race condition
        # https://github.com/fedora-infra/bodhi/issues/458
        time.sleep(1)

        if not alias:
            log.error("Update Handler got update with no "
                      "alias %s." % pprint.pformat(msg))
            return

        with self.db_factory() as session:
            update = Update.get(alias)
            if not update:
                raise BodhiException("Couldn't find alias '%s' in DB" % alias)

            if topic.endswith('update.edit'):
                bugs = [Bug.get(idx) for idx in msg['new_bugs']]
                # Sanity check
                for bug in bugs:
                    assert bug in update.bugs
            elif topic.endswith('update.request.testing'):
                bugs = update.bugs
            else:
                raise NotImplementedError("Should never get here.")

            self.work_on_bugs(session, update, bugs)
            self.fetch_test_cases(session, update)

        if config['test_gating.required']:
            with self.db_factory() as session:
                update = Update.get(alias)
                update.update_test_gating_status()

        log.info("Updates Handler done with %s, %s" % (alias, topic))

    def fetch_test_cases(self, session, update):
        """
        Query the wiki for test cases for each package on the given update.

        Args:
            session (sqlalchemy.orm.session.Session): A database session.
            update (bodhi.server.models.Update): The update's builds are iterated upon to find test
                cases for their associated Packages..
        """
        for build in update.builds:
            build.package.fetch_test_cases(session)

    def work_on_bugs(self, session, update, bugs):
        """
        Iterate the list of bugs, retrieving information from Bugzilla and modifying them.

        Iterate the given list of bugs associated with the given update. For each bug, retrieve
        details from Bugzilla, comment on the bug to let watchers know about the update, and mark
        the bug as MODIFIED. If the bug is a security issue, mark the update as a security update.

        If the bug is private, Bodhi can't retrieve any information, comment on it, or modify
        it, so we just associate the bug id with the update and mark it to be private.

        If handle_bugs is not True, return and do nothing.

        Args:
            session (sqlalchemy.orm.session.Session): A database session.
            update (bodhi.server.models.Update): The update that the bugs are associated with.
            bugs (list): A list of bodhi.server.models.Bug instances that we wish to act on.
        """
        if not self.handle_bugs:
            log.warning("Not configured to handle bugs")
            return

        log.info("Got %i bugs to sync for %r" % (len(bugs), update.alias))
        for bug in bugs:
            log.info("Getting RHBZ bug %r" % bug.bug_id)
            try:
                rhbz_bug = bug_module.bugtracker.getbug(bug.bug_id)

                log.info("Updating our details for %r" % bug.bug_id)
                bug.update_details(rhbz_bug)
                if bug.private:
                    # Bodhi can't retrieve any information so just continue with the next bug
                    log.info("  Skipping bug %r because it is private" % (bug.bug_id))
                    continue
                log.info("  Got title %r for %r" % (bug.title, bug.bug_id))

                # If you set the type of your update to 'enhancement' but you
                # attach a security bug, we automatically change the type of your
                # update to 'security'. We need to do this first, so we don't
                # accidentally comment on stuff that we shouldn't.
                if bug.security:
                    log.info("Setting our UpdateType to security.")
                    update.type = UpdateType.security

                log.info("Commenting on %r" % bug.bug_id)
                comment = config['initial_bug_msg'] % (
                    update.title, update.release.long_name, update.abs_url())

                log.info("Modifying %r" % bug.bug_id)
                bug.modified(update, comment)
            except Exception:
                log.warning('Error occurred during updating single bug', exc_info=True)



# -*- coding: utf-8 -*-
# Copyright © 2016 Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Drop the Update.karma column.

Revision ID: 06aa0e8aa5d2
Revises: 3cde3882442a
Create Date: 2016-10-26 16:55:54.875994
"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '06aa0e8aa5d2'
down_revision = None


def upgrade():
    """Drop the karma column from the updates table."""
    op.drop_column('updates', 'karma')


def downgrade():
    """
    Downgrade is not supported.

    If we ever do want to do this for some reason, we can use the
    code from the bodhi.server.models.Update.karma() property that was written in the same
    commit that introduced this migration as a guide for how to calculate the karma column. As it is
    highly unlikely that will ever be needed, this function simply raises an Exception for now.

    Raises:
        NotImplemented: This is raised if this function is executed.
    """
    raise NotImplementedError('Downgrade not supported')



# -*- coding: utf-8 -*-
# Copyright © 2014-2017 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Create and manage releases in Bodhi.

Here is an example of creatig a release:

    managereleases.py create --name F23 --long-name "Fedora 23" --id-prefix FEDORA --version 23
    --branch f23 --dist-tag f23 --stable-tag f23-updates --testing-tag f23-updates-testing
    --candidate-tag f23-updates-candidate --pending-stable-tag f23-updates-pending
    --pending-testing-tag f23-updates-testing-pending --override-tag f23-override --state pending
"""

import click

from bodhi.client import cli


def main():
    """
    Create and manage releases in Bodhi.

    (This utility has been deprecated. Please use 'bodhi releases' instead.)
    """
    click.echo("This utility has been deprecated. Please use 'bodhi releases' instead.")
    cli.commands['releases'].commands['create'].params[0].opts = ['--username']
    cli.commands['releases'].commands['edit'].params[0].opts = ['--username']
    cli.commands['releases']()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright © 2018 Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Define Bodhi's WSGI application.

As of the writing of this docblock, this module is a bit misnamed since the webapp is actually
defined in bodhi.server.__init__. However, that is an anti-pattern with lots of nasty in-line
imports due to circular dependencies, and this module is intended to solve that problem.
Unfortunately, it is a backwards-incompatible change to move main() here, so it will remain in
__init__ until we make a major Bodhi release. See https://github.com/fedora-infra/bodhi/issues/2294
"""

from pyramid.events import NewRequest, subscriber

from bodhi import server


def _complete_database_session(request):
    """
    Commit the database changes if no exceptions occurred.

    This is a post-request hook. It handles rolling back or committing the session based on whether
    an exception occurred or not. To get a database session that's not tied to the request/response
    cycle, just use the :data:`Session` scoped session.

    Args:
        request (pyramid.request.Request): The current web request.
    """
    _rollback_or_commit(request)
    server.Session().close()
    server.Session.remove()


@subscriber(NewRequest)
def _prepare_request(event):
    """
    Prepare each incoming request to Bodhi.

    This function does two things:
        * If requests do not have an Accept header, or if their Accept header is "*/*", it sets the
          header to application/json. Pyramid has undefined behavior when an ambiguous or missing
          Accept header is received, and multiple views are defined that handle specific Accept
          headers. For example, we have a view that returns html or JSON for /composes/, depending
          on the Accept header, but if a request has no Accept header or has */*, Pyramid will
          consider both views to be a match for the request and so it is undefined which view will
          handle the request. Let's force ambibuous requests to receive a JSON response so we have a
          defined behavior. See https://github.com/fedora-infra/bodhi/issues/2731.
        * It adds a callback to clean up the database session when the request is finished.

    Args:
        event (pyramid.events.NewRequest): The new request event.
    """
    if 'Accept' not in event.request.headers or event.request.headers['Accept'] == '*/*':
        event.request.headers['Accept'] = 'application/json'

    event.request.add_finished_callback(_complete_database_session)


def _rollback_or_commit(request):
    """
    Commit the transaction if there are no exceptions, otherwise rollback.

    Args:
        request (pyramid.request.Request): The current web request.
    """
    if request.exception is not None:
        server.Session().rollback()
    else:
        server.Session().commit()



# -*- coding: utf-8 -*-
# Copyright © 2007-2018 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Random functions that don't fit elsewhere."""

from collections import defaultdict, OrderedDict
from contextlib import contextmanager
import functools
import hashlib
import json
import os
import pkg_resources
import re
import socket
import subprocess
import tempfile
import time

from kitchen.iterutils import iterate
from pyramid.i18n import TranslationStringFactory
import arrow
import bleach
import colander
import hawkey
import libcomps
import libravatar
import librepo
import markdown
import requests
import rpm
from six.moves import map
from six.moves.urllib.parse import urlencode
import six

from bodhi.server import ffmarkdown, log, buildsys, Session
from bodhi.server.config import config
from bodhi.server.exceptions import RepodataException


_ = TranslationStringFactory('bodhi')

http_session = requests.Session()


def header(x):
    """Display a given message as a heading."""
    return u"%s\n     %s\n%s\n" % ('=' * 80, x, '=' * 80)


def get_rpm_header(nvr, tries=0):
    """
    Get the rpm header for a given build.

    Args:
        nvr (basestring): The name-version-release string of the build you want headers for.
        tries (int): The number of attempts that have been made to retrieve the nvr so far. Defaults
            to 0.
    Returns:
        dict: A dictionary mapping RPM header names to their values, as returned by the Koji client.
    """
    tries += 1
    headers = [
        'name', 'summary', 'version', 'release', 'url', 'description',
        'changelogtime', 'changelogname', 'changelogtext',
    ]
    rpmID = nvr + '.src'
    koji_session = buildsys.get_session()
    try:
        result = koji_session.getRPMHeaders(rpmID=rpmID, headers=headers)
    except Exception as e:
        msg = "Failed %i times to get rpm header data from koji for %s:  %s"
        log.warning(msg % (tries, nvr, str(e)))
        if tries < 3:
            # Try again...
            return get_rpm_header(nvr, tries=tries)
        else:
            # Give up for good and re-raise the failure...
            raise

    if result:
        return result

    raise ValueError("No rpm headers found in koji for %r" % nvr)


def flash_log(msg):
    """
    Log the given message at debug level.

    Args:
        msg (basestring): The message to log.
    """
    log.debug(msg)


def build_evr(build):
    """
    Return a tuple of strings of the given build's epoch, version, and release.

    Args:
        build (dict): A dictionary representing a Koji build with keys 'epoch', 'version', and
            'release'.
    Returns:
        tuple: A 3-tuple of strings representing the given build's epoch, version, and release,
            respectively.
    """
    if not build['epoch']:
        build['epoch'] = 0
    return tuple(map(str, (build['epoch'], build['version'], build['release'])))


def link(href, text):
    """
    Return an HTML anchor tag for the given href using the given text.

    Args:
        href (basestring): The URL to link.
        text (basestring): The text to render for the link.
    Returns:
        basestring: The requested anchor tag.
    """
    return '<a href="%s">%s</a>' % (href, text)


class memoized(object):
    """Decorator that permanently caches a function's return value each time it is called.

    If the function is called later with the same arguments, the cached value is returned (not
    reevaluated).

    http://wiki.python.org/moin/PythonDecoratorLibrary#Memoize

    Attributes:
        func (callable): The wrapped function.
        cache (dict): The cache, mapping arguments to the cached response.
    """

    def __init__(self, func):
        """
        Initialize the memoized object.

        Args:
            func: The function the memoized object is wrapping.
        """
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        """
        If the args are cached, return the cached value. If not, call the wrapped function.

        If the wrapped function is called, it's response is only cached if the args are hashable.

        Args:
            args (list): The list of arguments passed to the wrapped function.
        Returns:
            object: The reponse from the wrapped function, or the cached response, if available.
        """
        try:
            hash(args)
        except TypeError:
            # uncacheable. a list, for instance.
            # better to not cache than blow up.
            return self.func(*args)
        if args in self.cache:
            return self.cache[args]
        else:
            value = self.func(*args)
            self.cache[args] = value
            return value

    def __repr__(self):
        """
        Return the function's docstring.

        Returns:
            basestring: The wrapped function's docstring.
        """
        return self.func.__doc__

    def __get__(self, obj, objtype):
        """
        Support instance methods.

        Args:
            obj (object): The instance of the object the wrapped method is bound to.
            objtype (type): The type of the instance of the object the wrapped method is bound to.
        Returns:
            callable: A functools.partial response with the wrapped method's instance passed to it.
        """
        return functools.partial(self.__call__, obj)


@memoized
def get_critpath_components(collection='master', component_type='rpm', components=None):
    """
    Return a list of critical path packages for a given collection, filtered by components.

    Args:
        collection (basestring): The collection/branch to search. Defaults to 'master'.
        component_type (basestring): The component type to search for. This only affects PDC
            queries. Defaults to 'rpm'.
        components (frozenset or None): The list of components we are interested in. If None (the
            default), all components for the given collection and type are returned.
    Returns:
        list: The critpath components for the given collection and type.
    """
    critpath_components = []
    critpath_type = config.get('critpath.type')
    if critpath_type != 'pdc' and component_type != 'rpm':
        log.warning('The critpath.type of "{0}" does not support searching for'
                    ' non-RPM components'.format(component_type))

    if critpath_type == 'pkgdb':
        from pkgdb2client import PkgDB
        pkgdb = PkgDB(config.get('pkgdb_url'))
        results = pkgdb.get_critpath_packages(branches=collection)
        if collection in results['pkgs']:
            critpath_components = results['pkgs'][collection]
    elif critpath_type == 'pdc':
        critpath_components = get_critpath_components_from_pdc(
            collection, component_type, components)
    else:
        critpath_components = config.get('critpath_pkgs')

    # Filter the list of components down to what was requested, in case the specific path did
    # not take our request into account.
    if components is not None:
        critpath_components = [c for c in critpath_components if c in components]

    return critpath_components


def sanity_check_repodata(myurl, repo_type):
    """
    Sanity check the repodata for a given repository.

    Args:
        myurl (basestring): A path to a repodata directory.
        repo_type (str): This should be set to 'yum' for Yum repositories, 'module' for module
            repositories, or 'source' for source repositories.
    Raises:
        Exception: If the repodata is not valid or does not exist.
        ValueError: If repo_type is not an acceptable value.
    """
    if repo_type not in ('module', 'source', 'yum'):
        raise ValueError('repo_type must be one of module, source, or yum.')

    h = librepo.Handle()
    h.setopt(librepo.LRO_REPOTYPE, librepo.LR_YUMREPO)
    h.setopt(librepo.LRO_DESTDIR, tempfile.mkdtemp())

    if myurl[-1] != '/':
        myurl += '/'
    if myurl.endswith('repodata/'):
        myurl = myurl.replace('repodata/', '')

    h.setopt(librepo.LRO_URLS, [myurl])
    h.setopt(librepo.LRO_LOCAL, True)
    h.setopt(librepo.LRO_CHECKSUM, True)
    h.setopt(librepo.LRO_IGNOREMISSING, False)
    r = librepo.Result()
    try:
        h.perform(r)
    except librepo.LibrepoException as e:
        rc, msg, general_msg = e.args
        raise RepodataException(msg)

    repo_info = r.getinfo(librepo.LRR_YUM_REPO)
    primary_sack = hawkey.Sack()
    hk_repo = hawkey.Repo(myurl)
    try:
        hk_repo.filelists_fn = repo_info['filelists']
        # Source and module repos don't have DRPMs.
        if repo_type == 'yum':
            hk_repo.presto_fn = repo_info['prestodelta']
        elif repo_type == 'module':
            hk_repo.presto_fn = repo_info['modules']
        hk_repo.primary_fn = repo_info['primary']
        hk_repo.repomd_fn = repo_info['repomd']
        hk_repo.updateinfo_fn = repo_info['updateinfo']
    except KeyError as e:
        raise RepodataException('Required part not in repomd.xml: {}'.format(e.args[0]))
    primary_sack.load_repo(hk_repo,
                           build_cache=False,
                           load_filelists=True,
                           load_presto=True,
                           load_updateinfo=True)

    # Only yum repos have comps
    if repo_type == 'yum':
        # Test comps
        comps = libcomps.Comps()
        try:
            ret = comps.fromxml_f(repo_info['group'])
        except Exception:
            raise RepodataException('Comps file unable to be parsed')
        if len(comps.groups) < 1:
            raise RepodataException('Comps file empty')

    # Test updateinfo
    ret = subprocess.call(['zgrep', '<id/>', repo_info['updateinfo']])
    if not ret:
        raise RepodataException('updateinfo.xml.gz contains empty ID tags')


def age(context, date, nuke_ago=False):
    """
    Return a human readable age since the given date.

    Args:
        context (mako.runtime.Context): The current template rendering context.
        date (datetime.datetime): A date you wish a human readable age since.
        nuke_ago (bool): If True, remove " ago" from the age. Defaults to False.
    Returns:
        basestring: A human readable age since the given date.
    """
    humanized = arrow.get(date).humanize()
    if nuke_ago:
        return humanized.replace(' ago', '')
    else:
        return humanized


hardcoded_avatars = {
    'bodhi': 'https://apps.fedoraproject.org/img/icons/bodhi-{size}.png',
    # Taskotron may have a new logo at some point.  Check this out:
    # https://mashaleonova.wordpress.com/2015/08/18/a-logo-for-taskotron/
    # Ask tflink before actually putting this in place though.  we need
    # a nice small square version.  It'll look great!
    # In the meantime, we can use this temporary logo.
    'taskotron': 'https://apps.fedoraproject.org/img/icons/taskotron-{size}.png'
}


def avatar(context, username, size):
    """
    Return a URL of an avatar for the given username of the given size.

    Args:
        context (mako.runtime.Context): The current template rendering context.
        username (basestring): The username to return an avatar URL for.
        size (int): The size of the avatar you wish to retrieve, in unknown libravatar units.
    Returns:
        basestring: A URL to an avatar for the given username.
    """
    # Handle some system users
    # https://github.com/fedora-infra/bodhi/issues/308
    if username in hardcoded_avatars:
        return hardcoded_avatars[username].format(size=size)

    # context is a mako context object
    request = context['request']
    https = request.registry.settings.get('prefer_ssl')

    @request.cache.cache_on_arguments()
    def get_libravatar_url(openid, https, size):
        return libravatar.libravatar_url(
            openid=openid,
            https=https,
            size=size,
            default='retro',
        )

    def work(username, size):
        openid = "http://%s.id.fedoraproject.org/" % username
        if config.get('libravatar_enabled'):
            if config.get('libravatar_dns'):
                return get_libravatar_url(openid, https, size)
            else:
                query = urlencode({'s': size, 'd': 'retro'})
                hash = hashlib.sha256(openid.encode('utf-8')).hexdigest()
                template = "https://seccdn.libravatar.org/avatar/%s?%s"
                return template % (hash, query)

        return 'libravatar.org'

    return work(username, size)


def splitter(value):
    """
    Parse a string or list of comma or space delimited builds, returning a list of the values.

    Examples:
        >>> util.splitter('one,two,,three,')
        ['one', 'two', 'three']
        >>> util.splitter(['one,two,,three,,', 'four'])
        ['one', 'two', 'three', 'four']

    Args:
        value (basestring, colander.null, or iterable): The value to interpret as a list.
    Returns:
        list: A list of strings.
    """
    if value == colander.null:
        return

    items = []
    for v in iterate(value):
        if isinstance(v, six.string_types):
            for item in v.replace(',', ' ').split():
                items.append(item)

        elif v is not None:
            items.append(v)

    return items


def version(context=None):
    """
    Return the Bodhi server's version.

    Args:
        context (mako.runtime.Context or None): Unused. Defaults to None.
    Returns:
        basestring: The Bodhi server's version.
    """
    return pkg_resources.get_distribution('bodhi').version


def hostname(context=None):
    """
    Return the Bodhi server's hostname.

    Args:
        context (mako.runtime.Context or None): Unused. Defaults to None.
    Returns:
        basestring: The Bodhi server's hostname.
    """
    return socket.gethostname()


def markup(context, text):
    """
    Return HTML from a markdown string.

    Args:
        context (mako.runtime.Context): Unused.
        text (basestring): Markdown text to be converted to HTML.
    Returns:
        basestring: HTML representation of the markdown text.
    """
    # determine the major component of the bleach version installed.
    # this is similar to the approach that Pagure uses to determine the bleach version
    # https://pagure.io/pagure/pull-request/2269#request_diff
    bleach_major_v = int(bleach.__version__.split('.')[0])

    # the only difference in the bleach API that we use between v1 and v2 is
    # the formatting of the attributes parameter. Bleach 1 only allowed you
    # to specify attributes to be whitelisted for all whitelisted tags.
    # Bleach 2 requires you to specify the list of attributes whitelisted for
    # specific tags.
    if bleach_major_v >= 2:
        markdown_attrs = {
            "img": ["src", "alt", "title"],
            "a": ["href", "alt", "title"],
            "div": ["class"],
        }
    else:
        markdown_attrs = [
            "src", "href", "alt", "title", "class"
        ]

    markdown_tags = [
        "h1", "h2", "h3", "h4", "h5", "h6",
        "b", "i", "strong", "em", "tt",
        "p", "br",
        "span", "div", "blockquote", "code", "hr", "pre",
        "ul", "ol", "li", "dd", "dt",
        "img",
        "a",
    ]

    markdown_text = markdown.markdown(text, extensions=['markdown.extensions.fenced_code',
                                                        ffmarkdown.BodhiExtension()])

    # previously, we linkified text in ffmarkdown.py, but this was causing issues like #1721
    # so now we use the bleach linkifier to do this for us.
    markdown_text = bleach.linkify(markdown_text, parse_email=True)

    # previously, we used the Safe Mode in python-markdown to strip all HTML
    # tags. Safe Mode is deprecated, so we now use Bleach to sanitize all HTML
    # tags after running it through the markdown parser
    return bleach.clean(markdown_text, tags=markdown_tags, attributes=markdown_attrs)


def composestate2html(context, state):
    """
    Render the given UpdateStatus as a span containing text.

    Args:
        context (mako.runtime.Context): Unused.
        state (bodhi.server.models.ComposeState): The ComposeState to render as a span
            tag.
    Returns:
        basestring: An HTML span tag representing the ComposeState.
    """
    cls = {
        'requested': 'primary',
        'pending': 'primary',
        'initializing': 'warning',
        'updateinfo': 'warning',
        'punging': 'warning',
        'notifying': 'warning',
        'cleaning': 'warning',
        'success': 'success',
        'failed': 'danger',
    }[state.value]
    return "<span class='label label-%s'>%s</span>" % (cls, state.description)


def status2html(context, status):
    """
    Render the given UpdateStatus as a span containing text.

    Args:
        context (mako.runtime.Context): Unused.
        severity (bodhi.server.models.UpdateStatus): The UpdateStatus to render as a span
            tag.
    Returns:
        basestring: An HTML span tag representing the UpdateStatus.
    """
    status = six.text_type(status)
    cls = {
        'pending': 'primary',
        'testing': 'warning',
        'stable': 'success',
        'unpushed': 'danger',
        'obsolete': 'default',
        'processing': 'info',
    }[status]
    status_desc = {
        'pending': 'The update has not yet been pushed to the testing or stable repositories.',
        'testing': 'The package is in the testing repository for people to test.',
        'stable': 'The package has been released to the stable repository.',
        'unpushed': 'The update has been removed from testing.',
        'obsolete': 'The package has been obsoleted by a different update.',
        'processing': 'Unused.',
    }[status]

    return "<span class='text-muted' data-toggle='tooltip' title='%s'>" % (status_desc) \
        + "<span class='label label-%s'>%s</span>" % (cls, status) \
        + "</span>"


def greenwave_unsatisfied_requirements_html(context, update):
    """
    Return an html representation of the given update's greenwave_unsatisfied_requirements field.

    This function collates the unsatisfied requirements into a heirarchy of requirement type, then
    test name, and then a list of builds that fail to satisfy that requirement type for that test
    name.

    Args:
        context (mako.runtime.Context): Unused.
        update (bodhi.server.models.Update): The update to render unsatisfied requirements for.
    Returns:
        str: A human readable representation of the update's unsatisfied requirements.
    """
    if not update.greenwave_unsatisfied_requirements:
        return ''

    # The Greenwave data is in a list of dictionaries that describe unsatisfied requirements. Let's
    # collate that data into type of requirement (such as missing test, failed test, etc.) --> test
    # name --> list of nvrs that fail that test name. To do this, we'll build a dictionary of
    # dictionaries of lists, like this:
    #
    # {'test-result-missing': {'dist.rpmdeplint': ['bodhi-3.6.1-fc28']}}
    reqs = defaultdict(lambda: defaultdict(list))
    for r in json.loads(update.greenwave_unsatisfied_requirements):
        # The Greenwave response contains a lot of duplicate information. For now, let's ignore
        # anything that isn't a koji build. https://pagure.io/greenwave/issue/169
        if r['item'].get('type') == 'koji_build':
            reqs[r['type']][r['testcase']].append(r['item']['item'])
    # Now that we've built our data structure, let's render it into an HTML snippet that can be
    # included in the update page.
    html = '<div>Unsatisfied requirements:'
    for t in reqs.keys():
        html = html + '<div>{}: {}</div>'.format(
            t,
            ', '.join([' {} ({})'.format(testcase, ', '.join(reqs[t][testcase]))
                      for testcase in reqs[t].keys()]))
    html = html + '</div>'
    return html


def state2class(context, state):
    """
    Classify the given ReleaseState value.

    Args:
        context (mako.runtime.Context): Unused.
        state (bodhi.server.models.ReleaseState): The ReleaseState value you wish to classify.
    Returns:
        basestring: A string representing the classification of the given ReleaseState. Can return
            'danger', 'warning', 'success', or 'default active'.
    """
    state = six.text_type(state)
    cls = {
        'disabled': 'default active',
        'pending': 'warning',
        'current': 'success',
        'archived': 'danger'
    }
    return cls[state] if state in cls.keys() else 'default'


def type2color(context, t):
    """
    Return a color to render the given UpdateType with.

    Args:
        context (mako.runtime.Context): The current template rendering context.
        t (bodhi.server.models.UpdateType): The UpdateType you wish to choose a color for.
    Returns:
        basestring: A string in the format rgba(RED, GREEN, BLUE, ALPHA), where RED, GREEN, BLUE,
            and ALPHA are replaced with numerical values to represent a color.
    """
    t = six.text_type(t)
    cls = {
        'bugfix': 'rgba(150,180,205,0.5)',
        'security': 'rgba(205,150,180,0.5)',
        'newpackage': 'rgba(150,205,180,0.5)',
        'default': 'rgba(200,200,200,0.5)'
    }
    return cls[t] if t in cls.keys() else cls['default']


def state2html(context, state):
    """
    Render the given ReleaseState as an HTML span tag.

    Args:
        context (mako.runtime.Context): The current template rendering context.
        state (bodhi.server.models.ReleaseState): The ReleaseState you wish to render as HTML.
    Returns:
        basestring: An HTML rendering of the given ReleaseState.
    """
    state_class = state2class(context, state)
    return "<span class='label label-%s'>%s</span>" % (state_class, state)


def karma2class(context, karma, default='default'):
    """
    Classify the given karma value if possible, or return the default value.

    Args:
        context (mako.runtime.Context): Unused.
        karma (int): The karma value you wish to classify.
        default (basestring): The default value if karma is not within the range -2 <= karma <= 2.
    Returns:
        basestring: A string representing the classification of the given karma. Can return
            'danger', 'info', 'success', or the default value.
    """
    if karma and karma >= -2 and karma <= 2:
        return {
            -2: 'danger',
            -1: 'danger',
            0: 'info',
            1: 'success',
            2: 'success',
        }.get(karma)
    return default


def karma2html(context, karma):
    """
    Render the given karma as a span labeled for rendering, or as a list of td tags of the same.

    Args:
        context (mako.runtime.Context): The current template rendering context.
        karma (int or tuple): The karma value or values you wish to render a span tag or td tags
            for.
    Returns:
        basestring: An HTML span tag or a list of td tags representing the karma or karma tuple.
    """
    # Recurse if we are handle multiple karma values
    if isinstance(karma, tuple):
        return '</td><td>'.join([karma2html(context, item) for item in karma])

    cls = karma2class(context, karma, None)

    if not cls:
        if karma < -2:
            cls = 'danger'
        elif karma == 0:
            cls = 'primary'
        else:
            cls = 'success'

    if karma > 0:
        karma = "+%i" % karma
    else:
        karma = "%i" % karma

    return "<span class='label label-%s'>%s</span>" % (cls, karma)


def type2html(context, kind):
    """
    Render the given UpdateType as a span containing text.

    Args:
        context (mako.runtime.Context): Unused.
        severity (bodhi.server.models.UpdateType): The UpdateType to render as a span
            tag.
    Returns:
        basestring: An HTML span tag representing the UpdateType.
    """
    kind = six.text_type(kind)
    cls = {
        'security': 'danger',
        'bugfix': 'warning',
        'newpackage': 'primary',
        'enhancement': 'success',
    }.get(kind)

    return "<span class='label label-%s'>%s</span>" % (cls, kind)


def type2icon(context, kind):
    """
    Render the given UpdateType as a span containing an icon.

    Args:
        context (mako.runtime.Context): Unused.
        severity (bodhi.server.models.UpdateType): The UpdateType to render as a span
            tag.
    Returns:
        basestring: An HTML span tag representing the UpdateType.
    """
    kind = six.text_type(kind)
    cls = {
        'security': 'danger',
        'bugfix': 'warning',
        'newpackage': 'primary',
        'enhancement': 'success',
    }.get(kind)

    if kind[0].lower() in 'aeiou':
        kind_article = 'an'
    else:
        kind_article = 'a'

    fontawesome = {
        'security': 'fa-shield',
        'bugfix': 'fa-bug',
        'newpackage': 'fa-archive',
        'enhancement': 'fa-bolt',
    }.get(kind)

    span = ("<span class='label label-%s' data-toggle='tooltip' "
            "title='This is %s %s update'><i class='fa fa-fw %s'></i></span>")
    return span % (cls, kind_article, kind, fontawesome)


def severity2html(context, severity):
    """
    Render the given UpdateSeverity as an HTML span tag.

    Args:
        context (mako.runtime.Context): Unused.
        severity (bodhi.server.models.UpdateSeverity): The UpdateSeverity to render as a span
            tag.
    Returns:
        basestring: An HTML span tag representing the UpdateSeverity.
    """
    severity = six.text_type(severity)
    cls = {
        'urgent': 'danger',
        'high': 'warning',
        'medium': 'primary',
        'low': 'success',
        'unspecified': 'default',
    }.get(severity)

    return "<span class='label label-%s'>%s</span>" % (cls, severity)


def request2html(context, request):
    """
    Render the given UpdateRequest as an HTML span tag.

    Args:
        context (mako.runtime.Context): Unused.
        request (bodhi.server.models.UpdateRequest): The UpdateRequest to render as a span tag.
    Returns:
        basestring: An HTML span tag representing the UpdateRequest.
    """
    request = six.text_type(request)
    cls = {
        'unpush': 'danger',
        'obsolete': 'default',
        'testing': 'warning',
        'stable': 'success',
        'batched': 'success',
    }.get(request)

    return "<span class='label label-%s'>%s</span>" % (cls, request)


def update2html(context, update):
    """
    Return an HTML anchor tag for the given update.

    Args:
        context (mako.runtime.Context): The current template rendering context.
        update (bodhi.server.models.Update or dict): The Update or dictionary serialization of an
            Update that you wish to receive an HTML anchor tag for.
    Returns:
        basestring: An HTML anchor tag linking the Update.
    """
    request = context.get('request')

    if hasattr(update, 'title'):
        title = update.title
    else:
        title = update['title']

    if hasattr(update, 'alias'):
        alias = update.alias
    else:
        alias = update['alias']

    url = request.route_url('update', id=alias or title)
    settings = request.registry.settings
    max_length = settings.get('max_update_length_for_ui')
    if len(title) > max_length:
        title = title[:max_length] + "..."
    return link(url, title)


def pages_list(context, page, pages):
    """
    Return a list of page numbers to display in pagination lists.

    Args:
        context (mako.runtime.Context): Unused.
        page (int): The current page number.
        pages (int): The total number of pages.
    Returns:
        list: A list of page numbers that should be displayed to the user.
    """
    margin = 4
    num_pages = (2 * margin) + 1

    if page <= margin + 1:
        # Current `page` is close to the beginning of `pages`
        min_page = 1
        max_page = min(pages, num_pages)

    elif (pages - page) >= margin:
        min_page = max(page - margin, 1)
        max_page = min(page + margin, pages)

    else:
        # Current `page` is close to the end of `pages`
        max_page = min(pages, page + margin)
        min_page = max(max_page - (num_pages - 1), 1)

    return range(min_page, max_page + 1)


def page_url(context, page):
    """
    Return a paginated version of the request URL with the given page.

    Args:
        context (mako.runtime.Context): The current template context, used to get the current path
            URL.
        page (basestring): The requested page number.
    Returns:
        basestring: The current path appended with a GET query for the requested page.
    """
    request = context.get('request')
    params = dict(request.params)
    params['page'] = page
    return request.path_url + "?" + urlencode(params)


def bug_link(context, bug, short=False):
    """
    Form a URL to a given bugzilla bug.

    Args:
        context: Unused.
        short (bool): If False, includes the title of the bug in the response, or a spinner if the
            title is unknown. If True, only returns the anchor tag. Defaults to False.
    Returns:
        basestring: The requested link.
    """
    url = "https://bugzilla.redhat.com/show_bug.cgi?id=" + str(bug.bug_id)
    display = "#%i" % bug.bug_id
    link = "<a target='_blank' href='%s'>%s</a>" % (url, display)
    if not short:
        if bug.title:
            if bug.private:
                link = link + " <span class='label label-danger'>Private bug</span>"
            else:
                # We're good, but we do need to clean the bug title in case it contains malicious
                # tags. See CVE-2017-1002152: https://github.com/fedora-infra/bodhi/issues/1740
                link = link + " " + bleach.clean(bug.title, tags=[], attributes=[])
        else:
            # Otherwise, the backend is async grabbing the title from rhbz, so
            link = link + " <img class='spinner' src='static/img/spinner.gif'>"

    return link


def push_to_batched_or_stable_button(context, update):
    """
    Form the push to batched or push to stable button, as appropriate.

    Args:
        context (mako.runtime.Context): The current template rendering context. Unused.
        update (bodhi.server.models.Update): The Update we are rendering a button about.
    Returns:
        basestring: HTML for a button that will draw the push to stable or push to batched button,
            as appropriate.
    """
    if getattr(update.request, 'description', None) == 'batched' or \
            getattr(update.severity, 'description', None) == 'urgent':
        button = ('stable', 'Stable')
    elif getattr(update.request, 'description', None) in ('stable', None):
        button = ('batched', 'Batched')
    else:
        return ''

    return ('<a id="{}" class="btn btn-sm btn-success">'
            '<span class="fa fa-fw fa-arrow-circle-right"></span> Push to {}</a>').format(*button)


def testcase_link(context, test, short=False):
    """
    Form a URL to a given test description.

    Args:
        context: Unused.
        test (bodhi.server.models.TestCase): The test case you wish to have a link to.
        short (bool): If False, returns "Test Case " then the HTML anchor tag with the link. If
            True, only returns the anchor tag. Defaults to False.
    Returns:
        basestring: The requested link.
    """
    url = config.get('test_case_base_url') + test.name
    display = test.name.replace('QA:Testcase ', '')
    link = "<a target='_blank' href='%s'>%s</a>" % (url, display)
    if not short:
        link = "Test Case " + link
    return link


def can_waive_test_results(context, update):
    """
    Return True or False if the test results can be waived on an update.

    Args:
        context (mako.runtime.Context): The current template rendering context. Unused.
        update (bodhi.server.models.Update): The Update on which we are going to waive test results.
    Returns:
        bool: Indicating if the test results can be waived on the given update.
    """
    return config.get('test_gating.required') and not update.test_gating_passed \
        and config.get('waiverdb.access_token') and update.status.description != 'stable'


def sorted_builds(builds):
    """
    Sort the given builds by their NVRs.

    Args:
        builds (iterable): The builds you wish to sort by NVR.
    Returns:
        list: A list of Builds sorted by NVR.
    """
    key_function = functools.cmp_to_key(rpm.labelCompare)
    return sorted(builds,
                  key=lambda x: key_function(x.get_n_v_r()),
                  reverse=True)


def sorted_updates(updates):
    """
    Sort the given iterable of Updates so the highest version appears last.

    Order our updates so that the highest version gets tagged last so that
    it appears as the 'latest' in koji.

    Args:
        updates (iterable): An iterable of bodhi.server.models.Update objects to be sorted.
    Returns:
        tuple: A 2-tuple of lists. The first list contains builds that should be tagged
            synchronously in a specific order. The second list can be tagged asynchronously in koji
            with a multicall.
    """
    builds = defaultdict(set)
    sync, async_ = [], []
    for update in updates:
        for build in update.builds:
            builds[build.nvr_name].add(build)
    # The sorted here is so we actually have a way to test this
    # Otherwise, we would be depending on the way Python orders dict keys
    for package in sorted(builds.keys()):
        if len(builds[package]) > 1:
            log.debug(builds[package])
            for build in sorted_builds(builds[package])[::-1]:
                if build.update not in sync:
                    sync.append(build.update)
                if build.update in async_:
                    async_.remove(build.update)
        else:
            build = list(builds[package])[0]
            if build.update not in async_ and build.update not in sync:
                async_.append(build.update)
    log.info('sync = %s' % ([up.title for up in sync],))
    log.info('async_ = %s' % ([up.title for up in async_],))
    if not (len(set(sync) & set(async_)) == 0 and len(set(sync) | set(async_)) == len(updates)):
        # There should be absolutely no way to hit this code path, but let's be paranoid, and check
        # every run, to make sure no update gets left behind.
        # It makes sure that there is no update in sync AND async, and that the combination of
        # sync OR async_ is the full set of updates.
        raise Exception('ERROR! SYNC+ASYNC != UPDATES! sorted_updates failed')  # pragma: no cover
    return sync, async_


def cmd(cmd, cwd=None, raise_on_error=False):
    """
    Run the given command in a subprocess.

    Args:
        cmd (list): The command to be run. This is expressed as a list to be
            passed directly to subprocess.Popen().
        cwd (basestring or None): The current working directory to use when launching the
            subprocess.
        raise_on_error (bool): If True, raise a RuntimeError if the command's exit code is non-0.
            Defaults to False.
    Returns:
        tuple: A 3-tuple of the standard output (basestring), standard error (basestring), and the
            process's return code (int).
    Raises:
        RuntimeError: If exception is True and the command's exit code is non-0.
    """
    log.debug('Running {}'.format(' '.join(cmd)))
    p = subprocess.Popen(cmd, cwd=cwd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    output = '{}\n{}'.format(out, err)
    if p.returncode != 0:
        msg = '{} returned a non-0 exit code: {}'.format(' '.join(cmd), p.returncode)
        log.error(msg)
        log.error(output)
        if raise_on_error:
            raise RuntimeError(msg)
    elif out or err:
        log.debug(output)
    return out, err, p.returncode


def tokenize(string):
    """
    Interpret the given string as a space or comma separated ordered iterable of strings.

    For example, Given something like "a b, c d" return a generator equivalent of
    ['a', 'b', 'c', 'd'].

    Args:
        string (basestring): The string to be interpreted as an iterable.
    Yields:
        basestring: The individual tokens found in the comma or space separated string.
    """
    for substring in string.split(','):
        substring = substring.strip()
        if substring:
            for token in substring.split():
                token = token.strip()
                if token:
                    yield token


def taskotron_results(settings, entity='results/latest', max_queries=10, **kwargs):
    """
    Yield resultsdb results using query arguments.

    Args:
        entity (basestring): The API endpoint to use (see resultsdb documentation).
        max_queries (int): The maximum number of queries to perform (pages to retrieve). ``1`` means
            just a single page. ``None`` or ``0`` means no limit. Please note some tests might have
            thousands of results in the database and it's very reasonable to limit queries (thus the
            default value).
        kwargs (dict): Args that will be passed to resultsdb to specify what results to retrieve.
    Returns:
        generator or None: Yields Python objects loaded from ResultsDB's "data" field in its JSON
            response, or None if there was an Exception while performing the query.
    """
    max_queries = max_queries or 0
    url = settings['resultsdb_api_url'] + "/api/v2.0/" + entity
    if kwargs:
        url = url + "?" + urlencode(kwargs)
    data = True
    queries = 0

    try:
        while data and url:
            log.debug("Grabbing %r" % url)
            response = requests.get(url, timeout=60)
            if response.status_code != 200:
                raise IOError("status code was %r" % response.status_code)
            json = response.json()
            for datum in json['data']:
                yield datum

            url = json.get('next')
            queries += 1
            if max_queries and queries >= max_queries and url:
                log.debug('Too many result pages, aborting at: %r' % url)
                break
    except Exception as e:
        log.exception("Problem talking to %r : %r" % (url, str(e)))


class TransactionalSessionMaker(object):
    """Provide a transactional database scope around a series of operations."""

    @contextmanager
    def __call__(self):
        """
        Manage a database Session object for the life of the context.

        Yields a database Session object, then either commits the tranaction if there were no
        Exceptions or rolls back the transaction. In either case, it also will close and remove the
        Session.
        """
        session = Session()
        try:
            yield session
            session.commit()
        except Exception as e:
            # It is possible for session.rolback() to raise Exceptions, so we will wrap it in an
            # Exception handler as well so we can log the rollback failure and still raise the
            # original Exception.
            try:
                session.rollback()
            except Exception:
                log.exception('An Exception was raised while rolling back a transaction.')
            raise e
        finally:
            session.close()
            Session.remove()


transactional_session_maker = TransactionalSessionMaker


def sort_severity(value):
    """
    Map a given UpdateSeverity string representation to a numerical severity value.

    This is used to sort severities from low to high.

    Args:
        value (basestring): The human readable UpdateSeverity string.
    Returns:
        int: A number representing the sorting order of the given UpdateSeverity string, or 99 if an
            unknown value is given.
    """
    value_map = {
        'unspecified': 1,
        'low': 2,
        'medium': 3,
        'high': 4,
        'urgent': 5
    }

    return value_map.get(value, 99)


def severity_updateinfo_str(value):
    """
    Convert a severity level into one apt for inclusion in the updateinfo XML.

    These values are compatible with RHEL repodata to make it easier to write
    portable clients.

    Args:
        value (basestring): The Bodhi severity to be mapped to a repodata severity string.
    Returns:
        basestring: A severity string to be included in repodata.
    """
    severity_map = {
        'unspecified': "None",
        'low': "Low",
        'medium': "Moderate",
        'high': "Important",
        'urgent': "Critical",
    }
    return severity_map.get(value, "None")


# If we need to know about more components than this constant, we will just get the full
# list, rather than a query per package. This is because at some point, just going through
# paging becomes more performant than getting the page for every component.
PDC_CRITPATH_COMPONENTS_GETALL_LIMIT = 10


def get_critpath_components_from_pdc(branch, component_type='rpm', components=None):
    """
    Search PDC for critical path packages based on the specified branch.

    Args:
        branch (basestring): The branch name to search by.
        component_type (basestring): The component type to search by. Defaults to ``rpm``.
        components (frozenset or None): The list of components we are interested in. If None (the
            default), all components for the given branch and type are returned.
    Returns:
        list: Critical path package names.
    """
    pdc_api_url = '{}/rest_api/v1/component-branches/'.format(
        config.get('pdc_url').rstrip('/'))
    query_args = {
        'active': 'true',
        'critical_path': 'true',
        'name': branch,
        'page_size': 100,
        'type': component_type,
        'fields': 'global_component'
    }
    # Create ordered dictionary with sorted query args to be able to compare URLs
    query_args = OrderedDict(sorted(query_args.items(), key=lambda x: x[0]))

    critpath_pkgs_set = set()
    if components and len(components) < PDC_CRITPATH_COMPONENTS_GETALL_LIMIT:
        # Do a query for every single component
        for component in components:
            query_args['global_component'] = component
            pdc_api_url_with_args = '{0}?{1}'.format(pdc_api_url, urlencode(query_args))
            pdc_request_json = pdc_api_get(pdc_api_url_with_args)
            for branch_rv in pdc_request_json['results']:
                critpath_pkgs_set.add(branch_rv['global_component'])
            if pdc_request_json['next']:
                raise Exception('We got paging when requesting a single component?!')
    else:
        pdc_api_url_with_args = '{0}?{1}'.format(pdc_api_url, urlencode(query_args))
        while True:
            pdc_request_json = pdc_api_get(pdc_api_url_with_args)

            for branch_rv in pdc_request_json['results']:
                critpath_pkgs_set.add(branch_rv['global_component'])

            if pdc_request_json['next']:
                pdc_api_url_with_args = pdc_request_json['next']
            else:
                # There are no more results to iterate through
                break
    return list(critpath_pkgs_set)


def call_api(api_url, service_name, error_key=None, method='GET', data=None, headers=None,
             retries=0):
    """
    Perform an HTTP request with response type and error handling.

    Args:
        api_url (basestring): The URL to query.
        service_name (basestring): The service name being queried (used to form human friendly error
            messages).
        error_key (basestring): The key that indexes error messages in the JSON body for the given
            service. If this is set to None, the JSON response will be used as the error message.
        method (basestring): The HTTP method to use for the request. Defaults to ``GET``.
        data (dict): Query string parameters that will be sent along with the request to the server.
        retries (int): The number of times to retry, each after a 1 second sleep, if we get a
            non-200 HTTP code. Defaults to 3.
    Returns:
        dict: A dictionary representing the JSON response from the remote service.
    Raises:
        RuntimeError: If the server did not give us a 200 code.
    """
    if data is None:
        data = dict()
    if method == 'POST':
        if headers is None:
            headers = {'Content-Type': 'application/json'}
        base_error_msg = (
            'Bodhi failed to send POST request to {0} at the following URL '
            '"{1}". The status code was "{2}".')
        rv = http_session.post(api_url,
                               headers=headers,
                               data=json.dumps(data),
                               timeout=60)
    else:
        base_error_msg = (
            'Bodhi failed to get a resource from {0} at the following URL '
            '"{1}". The status code was "{2}".')
        rv = http_session.get(api_url, timeout=60)

    if rv.status_code >= 200 and rv.status_code < 300:
        return rv.json()
    elif retries:
        time.sleep(1)
        return call_api(api_url, service_name, error_key, method, data, headers, retries - 1)
    elif rv.status_code == 500:
        # There will be no JSON with an error message here
        error_msg = base_error_msg.format(
            service_name, api_url, rv.status_code)
        log.error(error_msg)
        raise RuntimeError(error_msg)
    else:
        # If it's not a 500 error, we can assume that the API returned an error
        # message in JSON that we can log
        try:
            rv_error = rv.json()
            if error_key is not None:
                rv_error = rv_error.get(error_key)
        except ValueError:
            rv_error = ''
        error_msg = base_error_msg.format(
            service_name, api_url, rv.status_code)
        error_msg = '{0} The error was "{1}".'.format(error_msg, rv_error)
        log.error(error_msg)
        raise RuntimeError(error_msg)


def pagure_api_get(pagure_api_url):
    """
    Perform a GET request against Pagure.

    Args:
        pagure_api_url (basestring): The URL to GET, including query parameters.
    Returns:
        dict: A dictionary response representing the API response's JSON.
    Raises:
        RuntimeError: If the server did not give us a 200 code.
    """
    return call_api(pagure_api_url, service_name='Pagure', error_key='error', retries=3)


def pdc_api_get(pdc_api_url):
    """
    Perform a GET request against PDC.

    Args:
        pdc_api_url (basestring): The URL to GET, including query parameters.
    Returns:
        dict: A dictionary response representing the API response's JSON.
    Raises:
        RuntimeError: If the server did not give us a 200 code.
    """
    # There is no error_key specified because the error key is not consistent
    # based on the error message
    return call_api(pdc_api_url, service_name='PDC', retries=3)


def greenwave_api_post(greenwave_api_url, data):
    """
    Post a request to Greenwave.

    Args:
        greenwave_api_url (basestring): The URL to query.
        data (dict): The parameters to send along with the request.
    Returns:
        dict: A dictionary response representing the API response's JSON.
    Raises:
        RuntimeError: If the server did not give us a 200 code.
    """
    # There is no error_key specified because the error key is not consistent
    # based on the error message
    return call_api(greenwave_api_url, service_name='Greenwave', method='POST',
                    data=data, retries=3)


def waiverdb_api_post(waiverdb_api_url, data):
    """
    Post a request to WaiverDB.

    Args:
        waiverdb_api_url (basestring): The URL to query.
        data (dict): The parameters to send along with the request.
    Returns:
        dict: A dictionary response representing the API response's JSON.
    Raises:
        RuntimeError: If the server did not give us a 200 code.
    """
    # There is no error_key specified because the error key is not consistent
    # based on the error message
    return call_api(waiverdb_api_url, service_name='WaiverDB', method='POST',
                    data=data, headers={
                        'Content-Type': 'application/json',
                        'Authorization': 'Bearer %s' % config.get('waiverdb.access_token')
                    })


class no_autoflush(object):
    """
    A content manager that disables sqlalchemy's autoflush, restoring it afterwards.

    Adapted from https://bitbucket.org/zzzeek/sqlalchemy/wiki/UsageRecipes/DisableAutoflush
    """

    def __init__(self, session):
        """
        Store the session and remember its entrant state.

        Args:
            session (sqlalchemy.orm.session.Session): The session to disable autoflush on.
        """
        self.session = session
        self.autoflush = session.autoflush

    def __enter__(self):
        """Disable autoflush."""
        self.session.autoflush = False

    def __exit__(self, *args, **kwargs):
        """Restore autoflush to its entrant state. Args unused."""
        self.session.autoflush = self.autoflush


def _get_build_repository(build):
    """
    Return the registry repository name for the given Build from the container's pull string.

    Examples -
    'candidate-registry.fedoraproject.org/f29/cockpit:176-5.fc28' => 'f29/cockpit'.
    'candidate-registry.fedoraproject.org/myrepo@sha256:<hash>' => 'myrepo'.

    Args:
        build (bodhi.server.models.Build): A Build representing a container or flatpak.
    Returns:
        str: The registry repository name for the build.
    """
    koji = buildsys.get_session()
    koji_build = koji.getBuild(build.nvr)

    pull_specs = koji_build['extra']['image']['index']['pull']
    # All the pull specs should have the same repository, so which one we use is arbitrary
    base, tag = re.compile(r'[:@]').split(pull_specs[0], 1)
    server, repository = base.split('/', 1)

    return repository


def copy_container(build, destination_registry=None, destination_tag=None):
    """
    Copy a ContainerBuild from the source registry to a destination registry under the given tag.

    Args:
        build (bodhi.server.models.ContainerBuild): The build you wish to copy from the source tag
            to the destination tag.
        destination_registry (str or None): The registry to copy the build into. If None (the
            default), the container.destination_registry setting is used.
        destination_tag (str or None): The destination tag you wish to copy the source image to. If
            None (the default), the build's version and release are used to form the destination
            tag.
    Raises:
        RuntimeError: If skopeo returns a non-0 exit code.
    """
    source_registry = config['container.source_registry']
    source_tag = '{}-{}'.format(build.nvr_version, build.nvr_release)

    if destination_tag is None:
        destination_tag = source_tag
    if destination_registry is None:
        destination_registry = config['container.destination_registry']

    repository = _get_build_repository(build)

    source_url = _container_image_url(source_registry, repository, source_tag)
    destination_url = _container_image_url(destination_registry, repository, destination_tag)

    skopeo_cmd = [
        config.get('skopeo.cmd'), 'copy', source_url, destination_url]
    if config.get('skopeo.extra_copy_flags'):
        for flag in reversed(config.get('skopeo.extra_copy_flags').split(',')):
            skopeo_cmd.insert(2, flag)
    cmd(skopeo_cmd, raise_on_error=True)


def _container_image_url(registry, repository, tag=None):
    """
    Return a URL suitable for use in Skopeo for copying or deleting container images.

    For example, this can return a URL like
    docker://registry.fedoraproject.org/f27/httpd:0-3.f27container if tag is None, or
    docker://registry.fedoraproject.org/f27/httpd:build is tag is 'build'.

    Args:
        repository (str): the repository name in the registry, extracted from Koji metadata
        registry (str): The registry you want a URL for.
        tag (str or None): The tag in the registry you want to reference.
    Returns:
        str: A URL referencing the given build and tag in the given registry.
    """
    return 'docker://{}/{}:{}'.format(registry, repository, tag)


def get_absolute_path(location):
    """
    Return an absolute path in filesystem for a relative path value.

    For example, location 'bodhi:server/templates' can return something like:
        /home/bodhi/bodhi/server/templates/

    Args:
        location (str): The relative path you want to convert.
    Returns:
        str: An absolute path in the filesystem referencing to the given directory.
    """
    module, final = location.split(':')
    base = os.path.dirname(__import__(module).__file__)
    return base + "/" + final


def update_install_command(context, update):
    """
    Return the dnf command for installing the Update.

    Args:
        context (mako.runtime.Context): Unused.
        update (bodhi.server.models.Update): The Update you want to install.
    Returns:
        basestring: The dnf command to install the Update.
    """
    status = six.text_type(update.status)
    alias = update.alias
    update_type = six.text_type(update.type)

    if status != 'stable' and status != 'testing':
        raise ValueError('Only updates in stable or testing can be installed!')

    command = 'sudo dnf {}{} --advisory={}{}'.format(
        'install' if update_type == 'newpackage' else 'upgrade',
        ' --enablerepo=updates-testing' if status == 'testing' else '',
        alias,
        r' \*' if update_type == 'newpackage' else '')
    return command



# Copyright (c) 2018 Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Add a greenwave_unsatisfied_requirements column to the updates table.

Revision ID: 59c0f5fbc1b2
Revises: c21dd18b161a
Create Date: 2018-05-01 15:37:07.346034
"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '59c0f5fbc1b2'
down_revision = 'c21dd18b161a'


def upgrade():
    """Add a greenwave_unsatisfied_requirements to the updates table."""
    op.add_column('updates',
                  sa.Column('greenwave_unsatisfied_requirements', sa.UnicodeText(), nullable=True))


def downgrade():
    """Drop the greenwave_unsatisfied_requirements from the updates table."""
    op.drop_column('updates', 'greenwave_unsatisfied_requirements')



# -*- coding: utf-8 -*-
# Copyright © 2016-2018 Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
The "signed handler".

This module is responsible for marking builds as "signed" when they get moved
from the pending-signing to pending-updates-testing tag by RoboSignatory.
"""

import logging
import pprint

import fedmsg.consumers

from bodhi.server import initialize_db
from bodhi.server.config import config
from bodhi.server.models import Build
from bodhi.server.util import transactional_session_maker


log = logging.getLogger('bodhi')


class SignedHandler(fedmsg.consumers.FedmsgConsumer):
    """
    The Bodhi Signed Handler.

    A fedmsg listener waiting for messages from koji about builds being tagged.
    """

    config_key = 'signed_handler'

    def __init__(self, hub, *args, **kwargs):
        """
        Initialize the SignedHandler, configuring its topic and database.

        Args:
            hub (moksha.hub.hub.CentralMokshaHub): The hub this handler is consuming messages from.
                It is used to look up the hub config.
        """
        initialize_db(config)
        self.db_factory = transactional_session_maker()

        prefix = hub.config.get('topic_prefix')
        env = hub.config.get('environment')
        self.topic = [
            prefix + '.' + env + '.buildsys.tag'
        ]

        super(SignedHandler, self).__init__(hub, *args, **kwargs)
        log.info('Bodhi signed handler listening on:\n'
                 '%s' % pprint.pformat(self.topic))

    def consume(self, message):
        """
        Handle fedmsgs arriving with the configured topic.

        This marks a build as signed if it is assigned to the pending testing release tag.

        Example message format::
            {
                u'body': {
                    u'i': 628,
                    u'timestamp': 1484692585,
                    u'msg_id': u'2017-821031da-be3a-4f4b-91df-0baa834ca8a4',
                    u'crypto': u'x509',
                    u'topic': u'org.fedoraproject.prod.buildsys.tag',
                    u'signature': u'100% real please trust me',
                    u'msg': {
                        u'build_id': 442562,
                        u'name': u'colord',
                        u'tag_id': 214,
                        u'instance': u's390',
                        u'tag': 'f26-updates-testing-pending',
                        u'user': u'sharkcz',
                        u'version': u'1.3.4',
                        u'owner': u'sharkcz',
                        u'release': u'1.fc26'
                    },
                },
            }

        The message can contain additional keys.

        Args:
            message (dict): The incoming fedmsg in the format described above.
        """
        msg = message['body']['msg']

        build_nvr = '%(name)s-%(version)s-%(release)s' % msg
        tag = msg['tag']

        log.info("%s tagged into %s" % (build_nvr, tag))

        with self.db_factory():
            build = Build.get(build_nvr)
            if not build:
                log.info("Build was not submitted, skipping")
                return

            if not build.release:
                log.info('Build is not assigned to release, skipping')
                return

            if build.release.pending_testing_tag != tag:
                log.info("Tag is not pending_testing tag, skipping")
                return

            # This build was moved into the pending_testing tag for the applicable release, which
            # is done by RoboSignatory to indicate that the build has been correctly signed and
            # written out. Mark it as such.
            log.info("Build has been signed, marking")
            build.signed = True
            log.info("Build %s has been marked as signed" % build_nvr)



"""Add a scm_url column to the builds table.

Revision ID: 0e105f6e36b4
Revises: 8eaacb38b036
Create Date: 2017-05-11 18:35:40.550295
"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0e105f6e36b4'
down_revision = '8eaacb38b036'


def upgrade():
    """Add a scm_url column to the builds table with a uniqueness constraint."""
    op.add_column('builds', sa.Column('scm_url', sa.Unicode(length=256), nullable=True))
    op.create_unique_constraint('uq_scm_url', 'builds', ['scm_url'])


def downgrade():
    """Drop the scm_url column from the builds table and drop its uniqueness constraint."""
    op.drop_constraint('uq_scm_url', 'builds', type_='unique')
    op.drop_column('builds', 'scm_url')



# Copyright (c) 2017 Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Add packages type constraint.

Revision ID: 74cfbc6116ad
Revises: 95ce24bed77a
Create Date: 2017-11-21 19:52:30.058663
"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '74cfbc6116ad'
down_revision = '95ce24bed77a'


def upgrade():
    """Create a uniqueness constraint on packages.{name,type}."""
    op.create_unique_constraint('packages_name_and_type_key', 'packages', ['name', 'type'])


def downgrade():
    """Drop the uniqueness constraint on packages.{name,type}."""
    op.drop_constraint('packages_name_and_type_key', 'packages', type_='unique')



# -*- coding: utf-8 -*-
# Copyright 2014-2017 Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""A collection of fedmsg publishing utilities."""
import collections
import json
import logging
import socket

from sqlalchemy import event
from fedora_messaging import api, exceptions as fml_exceptions
import fedmsg
import fedmsg.config
import fedmsg.encoding

from bodhi.server import Session
import bodhi.server
import bodhi.server.config


_log = logging.getLogger(__name__)


def init(active=None, cert_prefix=None):
    """
    Initialize fedmsg for publishing.

    Args:
        active (bool or None): If True, publish messages to a relay. If False, publish messages to
            connected consumers.
        cert_prefix (basestring): Configures the ``cert_prefix`` setting in the fedmsg_config.
    """
    if not bodhi.server.config.config.get('fedmsg_enabled'):
        bodhi.server.log.warning("fedmsg disabled.  not initializing.")
        return

    fedmsg_config = fedmsg.config.load_config()

    # Only override config from disk if explicitly argued.
    if active is not None:
        fedmsg_config['active'] = active
        fedmsg_config['name'] = 'relay_inbound'
    else:
        hostname = socket.gethostname().split('.', 1)[0]
        fedmsg_config['name'] = 'bodhi.%s' % hostname

    if cert_prefix is not None:
        fedmsg_config['cert_prefix'] = cert_prefix

    fedmsg.init(**fedmsg_config)
    bodhi.server.log.info("fedmsg initialized")


@event.listens_for(Session, 'after_commit')
def send_fedmsgs_after_commit(session):
    """
    Send queued fedmsgs after a database commit.

    This relies on the session ``info`` dictionary being populated. At the moment,
    this is done by calling the :func:`publish` function. In the future it should
    be done automatically using SQLAlchemy listeners.

    Args:
        session (sqlalchemy.orm.session.Session): The session that was committed.
    """
    if 'fedmsg' in session.info:
        _log.debug('Emitting all queued fedmsgs for %r', session)
        # Initialize right before we try to publish, but only if we haven't
        # initialized for this thread already.
        if not fedmsg_is_initialized():
            init()

        for topic, messages in session.info['fedmsg'].items():
            _log.debug('emitting %d fedmsgs to the "%s" topic queued by %r',
                       len(messages), topic, session)
            for msg in messages:
                fedmsg.publish(topic=topic, msg=msg)
                _log.debug('Emitted a fedmsg, %r, on the "%s" topic, queued by %r',
                           msg, topic, session)
            # Tidy up after ourselves so a second call to commit on this session won't
            # send the same messages again. We cannot delete topic from fedmsg dict
            # because we cannot change dictionary size during iteration
            session.info['fedmsg'][topic] = []


@event.listens_for(Session, 'after_commit')
def send_messages_after_commit(session):
    """
    Send messages via AMQP after a database commit occurs.

    Args:
        session (sqlalchemy.orm.session.Session): The session that was committed.
    """
    if 'messages' in session.info:
        for m in session.info['messages']:
            try:
                api.publish(m)
            except fml_exceptions.BaseException:
                # In the future we should handle errors more gracefully
                _log.exception("An error occurred publishing %r after a database commit", m)
        session.info['messages'] = []


def publish(topic, msg, force=False):
    """
    Send a message via AMQP or fedmsg.

    This is used to send a message to fedmsg if "fedmsg_enabled" is true in the
    application configuration, otherwise it will use AMQP.

    Args:
        topic (str): The topic suffix. The "bodhi" prefix is applied, along with
            the "topic_prefix" and "environment" settings from fedmsg.
        msg (dict): The message body to send.
        force (bool): If False (the default), the message is only sent after the
            currently active database transaction successfully commits. If true,
            the messages is sent immediately.
    """
    if bodhi.server.config.config.get('fedmsg_enabled'):
        _fedmsg_publish(topic, msg, force)
        return

    # Dirty, nasty hack that I feel shame for: fedmsg modifies messages quietly
    # if they have objects with __json__ methods on them. For now, copy that
    # behavior. In the future, callers should pass fedora_messaging.api.Message
    # sub-classes or this whole API should go away.
    body = fedmsg.encoding.loads(fedmsg.encoding.dumps(msg))

    # Dirty, nasty hack #2: fedmsg mangles the topic in various ways. Duplicate
    # that behavior here for now. In the future callers should pass proper topics.
    fedmsg_conf = fedmsg.config.load_config()
    topic = "{prefix}.{env}.bodhi.{t}".format(
        prefix=fedmsg_conf["topic_prefix"], env=fedmsg_conf["environment"], t=topic)

    message = api.Message(topic=topic, body=body)
    if force:
        api.publish(message)
        return

    session = Session()
    if 'messages' not in session.info:
        session.info['messages'] = []
    session.info['messages'].append(message)
    _log.debug('Queuing message %r for delivery on session commit', message.id)


def _fedmsg_publish(topic, msg, force):
    """Publish a message to fedmsg.

    By default, messages are not sent immediately, but are queued in a
    transaction "data manager".  They will only get published after the
    sqlalchemy transaction completes successfully and will not be published at
    all if it fails, aborts, or rolls back.

    Specifying force=True to this function by-passes that -- messages are sent
    immediately.

    Args:
        topic (basestring): The message topic suffix.
        msg (dict): A dictionary representing the message to be published via fedmsg.
        force (bool): If True, send the message immediately. Else, queue the message to be sent
            after the current database transaction is committed.
    """
    # Initialize right before we try to publish, but only if we haven't
    # initialized for this thread already.
    if not fedmsg_is_initialized():
        init()

    if force:
        _log.debug("fedmsg skipping transaction and sending %r" % topic)
        fedmsg.publish(topic=topic, msg=msg)
    else:
        # We need to do this to ensure all the SQLAlchemy objects that could be in the messages
        # are turned into JSON before the session is removed and expires the objects loaded with
        # it. The JSON is decoded again because the fedmsg API doesn't state it accepts strings.
        # An issue has been filed about this: https://github.com/fedora-infra/fedmsg/issues/407.
        json_msg = fedmsg.encoding.dumps(msg)
        msg_dict = json.loads(json_msg)
        # This gives us the thread-local session which we'll use to stash the fedmsg.
        # When commit is called on it, the :func:`send_fedmsgs_after_commit` is triggered.
        session = Session()
        if 'fedmsg' not in session.info:
            _log.debug('Adding a dictionary for fedmsg storage to %r', session)
            session.info['fedmsg'] = collections.defaultdict(list)
        session.info['fedmsg'][topic].append(msg_dict)
        _log.debug('Enqueuing a fedmsg, %r, for topic "%s" on %r', msg_dict, topic, session)


def fedmsg_is_initialized():
    """
    Return True or False if fedmsg is initialized or not.

    Returns:
        bool: Indicating if fedmsg is initialized.
    """
    local = getattr(fedmsg, '__local')
    if not hasattr(local, '__context'):
        return False
    # Ensure that fedmsg has an endpoint to publish to.
    context = getattr(local, '__context')
    return hasattr(context, 'publisher')



# -*- coding: utf-8 -*-
# Copyright 2014-2018 Red Hat, Inc. and others
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Defines API services that pertain to users."""
import math

from cornice import Service
from cornice.validators import colander_querystring_validator
from pyramid.exceptions import HTTPNotFound
from sqlalchemy import func, distinct
from sqlalchemy.sql import or_

from bodhi.server.models import Group, Package, Update, User
from bodhi.server.validators import (validate_updates,
                                     validate_packages, validate_groups)
import bodhi.server.schemas
import bodhi.server.security
import bodhi.server.services.errors
import bodhi.server.services.updates


user = Service(name='user', path='/users/{name}',
               description='Bodhi users',
               # These we leave wide-open since these are only GETs
               cors_origins=bodhi.server.security.cors_origins_ro)

users = Service(name='users', path='/users/',
                description='Bodhi users',
                # These we leave wide-open since these are only GETs
                cors_origins=bodhi.server.security.cors_origins_ro)

users_rss = Service(name='users_rss', path='/rss/users/', description='Bodhi users RSS feed',
                    cors_origins=bodhi.server.security.cors_origins_ro)


@user.get(accept=("application/json", "text/json"), renderer="json",
          error_handler=bodhi.server.services.errors.json_handler)
@user.get(accept=("application/javascript"), renderer="jsonp",
          error_handler=bodhi.server.services.errors.json_handler)
@user.get(accept="text/html", renderer="user.html",
          error_handler=bodhi.server.services.errors.html_handler)
def get_user(request):
    """
    Return a user given by username.

    Args:
        request (pyramid.request): The current request.
    Returns:
        dict: A dictionary with two keys. "user" maps to a dictionary representation of the User
            object. "urls" maps to various URLs that describe various other objects related to the
            user.
    """
    id = request.matchdict.get('name')
    user = User.get(id)

    if not user:
        request.errors.add('body', 'name', 'No such user')
        request.errors.status = HTTPNotFound.code
        return

    user = user.__json__(request)

    # Throw some extra information in there
    rurl = request.route_url  # Just shorthand
    urls = {
        'comments_by': rurl('comments') + '?user=%s' % id,
        'comments_on': rurl('comments') + '?update_owner=%s' % id,
        'recent_updates': rurl('updates') + '?user=%s' % id,
        'recent_overrides': rurl('overrides') + '?user=%s' % id,
        'comments_by_rss': rurl('comments_rss') + '?user=%s' % id,
        'comments_on_rss': rurl('comments_rss') + '?update_owner=%s' % id,
        'recent_updates_rss': rurl('updates_rss') + '?user=%s' % id,
        'recent_overrides_rss': rurl('overrides_rss') + '?user=%s' % id,
    }

    return dict(user=user, urls=urls)


validators = (
    colander_querystring_validator,
    validate_groups,
    validate_updates,
    validate_packages,
)


@users.get(schema=bodhi.server.schemas.ListUserSchema,
           accept=("application/json", "text/json"), renderer="json",
           error_handler=bodhi.server.services.errors.json_handler,
           validators=validators)
@users.get(schema=bodhi.server.schemas.ListUserSchema,
           accept=("application/javascript"), renderer="jsonp",
           error_handler=bodhi.server.services.errors.jsonp_handler,
           validators=validators)
@users.get(schema=bodhi.server.schemas.ListUserSchema, renderer="rss",
           accept=('application/atom+xml',),
           error_handler=bodhi.server.services.errors.html_handler,
           validators=validators)
@users_rss.get(schema=bodhi.server.schemas.ListUserSchema, renderer="rss",
               error_handler=bodhi.server.services.errors.html_handler,
               validators=validators)
def query_users(request):
    """
    Search for users by various criteria.

    Args:
        request (pyramid.request): The current web request.
    Returns:
        dict: A dictionary with the follow key mappings:
            users: A list of users matching the search criteria.
            page: The current page of results.
            pages: The total number of pages available.
            rows_per_page: The number of users on the page.
            total: The total number of users matching the search criteria.
    """
    db = request.db
    data = request.validated
    query = db.query(User)

    like = data.get('like')
    if like is not None:
        query = query.filter(or_(*[
            User.name.like('%%%s%%' % like)
        ]))

    search = data.get('search')
    if search is not None:
        query = query.filter(User.name.ilike('%%%s%%' % search))

    name = data.get('name')
    if name is not None:
        query = query.filter(User.name.like(name))

    groups = data.get('groups')
    if groups is not None:
        query = query.join(User.groups)
        query = query.filter(or_(*[Group.id == grp.id for grp in groups]))

    updates = data.get('updates')
    if updates is not None:
        query = query.join(User.updates)
        args = \
            [Update.title == update.title for update in updates] +\
            [Update.alias == update.alias for update in updates]
        query = query.filter(or_(*args))

    packages = data.get('packages')
    if packages is not None:
        query = query.join(User.packages)
        query = query.filter(or_(*[Package.id == p.id for p in packages]))

    # We can't use ``query.count()`` here because it is naive with respect to
    # all the joins that we're doing above.
    count_query = query.with_labels().statement\
        .with_only_columns([func.count(distinct(User.id))])\
        .order_by(None)
    total = request.db.execute(count_query).scalar()

    page = data.get('page')
    rows_per_page = data.get('rows_per_page')
    pages = int(math.ceil(total / float(rows_per_page)))
    query = query.offset(rows_per_page * (page - 1)).limit(rows_per_page)

    return dict(
        users=query.all(),
        page=page,
        pages=pages,
        rows_per_page=rows_per_page,
        total=total,
    )



# -*- coding: utf-8 -*-
# Copyright © 2017 Caleigh Runge-Hottman and Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Add request for batched update.

Revision ID: 2a10629168e4
Revises: 0e105f6e36b4
Create Date: 2017-06-23 22:14:55.464629
"""
from alembic import op
from sqlalchemy import exc


# revision identifiers, used by Alembic.
revision = '2a10629168e4'
down_revision = '0e105f6e36b4'


def upgrade():
    """Add a 'batched' request enum value in ck_update_request."""
    op.execute('COMMIT')  # See https://bitbucket.org/zzzeek/alembic/issue/123
    try:
        # This will raise a ProgrammingError if the DB server doesn't use BDR.
        op.execute('SHOW bdr.permit_ddl_locking')
        # This server uses BDR, so let's ask for a DDL lock.
        op.execute('SET LOCAL bdr.permit_ddl_locking = true')
    except exc.ProgrammingError:
        # This server doesn't use BDR, so no problem.
        pass
    op.execute("ALTER TYPE ck_update_request ADD VALUE 'batched' AFTER 'stable'")


def downgrade():
    """Alert user that we cannot downgrade this migration."""
    raise NotImplementedError("Downgrading this migration is not supported.")



# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Contains the code that is used to generate the bodhi-expire-overrides server CLI."""

from datetime import datetime
import logging
import os
import sys

from pyramid.paster import get_appsettings, setup_logging

from ..buildsys import setup_buildsystem
from ..models import BuildrootOverride
from bodhi.server import Session, initialize_db


def usage(argv):
    """
    Print usage info and exit(1).

    Args:
        argv (list): The command line arguments that were passed to the CLI from sys.argv.
    """
    cmd = os.path.basename(argv[0])
    print('usage: %s <config_uri>\n'
          '(example: "%s development.ini")' % (cmd, cmd))
    sys.exit(1)


def main(argv=sys.argv):
    """
    Search for overrides that are past their expiration date and mark them expired.

    Args:
        argv (list): The command line arguments. Defaults to sys.argv.
    """
    if len(argv) != 2:
        usage(argv)

    config_uri = argv[1]

    setup_logging(config_uri)
    log = logging.getLogger(__name__)

    settings = get_appsettings(config_uri)
    initialize_db(settings)
    db = Session()

    setup_buildsystem(settings)

    try:
        now = datetime.utcnow()

        overrides = db.query(BuildrootOverride)
        overrides = overrides.filter(BuildrootOverride.expired_date.is_(None))
        overrides = overrides.filter(BuildrootOverride.expiration_date < now)

        count = overrides.count()

        if not count:
            log.info("No active buildroot override to expire")
            return

        log.info("Expiring %d buildroot overrides...", count)

        for override in overrides:
            override.expire()
            db.add(override)
            log.info("Expired %s" % override.build.nvr)
        db.commit()
    except Exception as e:
        log.error(e)
        db.rollback()
        Session.remove()
        sys.exit(1)



# -*- coding: utf-8 -*-
# Copyright © 2017 Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Remove the ci_status feature.

Revision ID: 95ce24bed77a
Revises: 33a416e03f50
Create Date: 2017-08-07 19:24:57.448750
"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '95ce24bed77a'
down_revision = '33a416e03f50'


def upgrade():
    """
    Remove database columns, enums, and constraints that were added as part of ci_status.

    Remove the ci_status enum and column, since ci_status has been superceded by test_gating_status.
    Then remove the scm_url field and its uniqueness constraint.
    """
    op.drop_column('builds', 'ci_status')
    op.execute("DROP TYPE ck_ci_status")
    op.drop_constraint('uq_scm_url', 'builds', type_='unique')
    op.drop_column('builds', 'scm_url')


def downgrade():
    """Restore the fields, enums, and constraints removed in the upgrade() function."""
    op.add_column('builds', sa.Column('scm_url', sa.Unicode(length=256), nullable=True))
    op.create_unique_constraint('uq_scm_url', 'builds', ['scm_url'])
    op.execute(
        "CREATE TYPE ck_ci_status AS ENUM "
        "('ignored', 'queued', 'running', 'passed', 'failed', 'waiting')")
    op.add_column(
        'builds',
        sa.Column(
            'ci_status',
            sa.Enum(
                'ignored', 'queued', 'running', 'passed', 'failed', 'waiting',
                name='ck_ci_status'),
            nullable=True
        )
    )



# Copyright (c) 2018 Sebastian Wojciechowski.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Add composed_by_bodhi flag to releases table.

Revision ID: d986618207bc
Revises: 3a14c47250fb
Create Date: 2018-12-11 21:52:14.081423
"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd986618207bc'
down_revision = '3a14c47250fb'


def upgrade():
    """Add a new composed_by_bodhi column to releases."""
    op.add_column(
        'releases', sa.Column('composed_by_bodhi', sa.Boolean(), server_default='t', default=True)
    )


def downgrade():
    """Drop the composed_by_bodhi column from the releases table."""
    op.drop_column('releases', 'composed_by_bodhi')



# -*- coding: utf-8 -*-
# Copyright © 2007-2017 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Define the /csrf services."""

from cornice import Service

import bodhi.server.security
import bodhi.server.services.errors


csrf = Service(name='csrf', path='/csrf', description='CSRF Token',
               # XXX - even though this is really a read-only endpoint,
               # we're going to **mark** it as read-write to CORS so
               # that somebody's exploited browser can't make a cross-site
               # request here to steal the CSRF token and escalate damage.
               cors_origins=bodhi.server.security.cors_origins_rw)


@csrf.get(accept="text/html", renderer="string",
          error_handler=bodhi.server.services.errors.html_handler)
def get_csrf_token_html(request):
    """
    Return a plain string CSRF token.

    Returns:
        basestring: A CSRF token.
    """
    return request.session.get_csrf_token()


@csrf.get(accept=("application/json", "text/json"), renderer="json",
          error_handler=bodhi.server.services.errors.json_handler)
def get_csrf_token_json(request):
    """
    Return a JSON string containing a CSRF token, in a key called csrf_token.

    Returns:
        basestring: A JSON string with a key csrf_token that references a CSRF token.
    """
    return dict(csrf_token=request.session.get_csrf_token())



"""
Used with the bodhi pshell.

    pshell /etc/bodhi/production.ini
    execfile('shelldb.py')
"""
from pyramid.paster import get_appsettings
import sys

from bodhi.server import Session, initialize_db


config_uri = None
for arg in sys.argv:
    if arg.endswith('.ini'):
        config_uri = arg
if not config_uri:
    config_uri = '/etc/bodhi/production.ini'


settings = get_appsettings(config_uri)
initialize_db(settings)
db = Session()


def delete_update(up):
        for b in up.builds:
                db.delete(b)
                if b.override:
                        db.delete(b.override)
        for c in up.comments:
                db.delete(c)
        db.delete(up)



# -*- coding: utf-8 -*-
# Copyright © 2013-2017 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Generates the bodhi-approve-testing console script.

The script is responsible for commenting on updates after they reach the mandatory amount of time
spent in the testing repository.
"""

import os
import sys
import logging

from pyramid.paster import get_appsettings
import six

from ..models import Update, UpdateStatus
from ..config import config
from bodhi.server import Session, initialize_db, notifications


logger = logging.getLogger('approve-testing')
logging.basicConfig(level=logging.INFO)


def usage(argv):
    """
    Print usage information and exit with code 1.

    Args:
        argv (list): The arguments that were passed to the CLI from sys.argv.
    """
    cmd = os.path.basename(argv[0])
    print('usage: %s <config_uri>\n'
          '(example: "%s development.ini")' % (cmd, cmd))
    sys.exit(1)


def main(argv=sys.argv):
    """
    Comment on updates that are eligible to be pushed to stable.

    Queries for updates in the testing state that have a NULL request, looping over them looking for
    updates that are eligible to be pushed to stable but haven't had comments from Bodhi to this
    effect. For each such update it finds it will add a comment stating that the update may now be
    pushed to stable.

    This function is the entry point for the bodhi-approve-testing console script.

    Args:
        argv (list): A list of command line arguments. Defaults to sys.argv.
    """
    if len(argv) != 2:
        usage(argv)

    settings = get_appsettings(argv[1])
    initialize_db(settings)
    db = Session()

    try:
        testing = db.query(Update).filter_by(status=UpdateStatus.testing,
                                             request=None)
        for update in testing:
            # If this release does not have any testing requirements, skip it
            if not update.release.mandatory_days_in_testing:
                print('%s doesn\'t have mandatory days in testing' % update.release.name)
                continue

            # If this has already met testing requirements, skip it
            if update.met_testing_requirements:
                continue

            # Approval message when testing based on karma threshold
            if update.stable_karma not in (0, None) and update.karma >= update.stable_karma \
                    and not update.autokarma and update.meets_testing_requirements:
                print('%s now reaches stable karma threshold' % update.title)
                text = config.get('testing_approval_msg_based_on_karma')
                update.comment(db, text, author=u'bodhi')
                db.commit()
                continue

            # If autokarma updates have reached the testing threshold, say something! Keep in mind
            # that we don't care about karma here, because autokarma updates get their request set
            # to stable by the Update.comment() workflow when they hit the required threshold. Thus,
            # this function only needs to consider the time requirements because these updates have
            # not reached the karma threshold.
            if update.meets_testing_requirements:
                print('%s now meets testing requirements' % update.title)
                text = six.text_type(
                    config.get('testing_approval_msg') % update.mandatory_days_in_testing)
                update.comment(db, text, author=u'bodhi')

                notifications.publish(
                    topic='update.requirements_met.stable',
                    msg=dict(update=update))
                db.commit()

    except Exception as e:
        print(str(e))
        db.rollback()
        Session.remove()
        sys.exit(1)



"""Convert the builds table to be polymorphic.

Revision ID: 9241378c92ab
Revises: 12d3e8695f90
Create Date: 2017-04-06 20:37:24.766366
"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '9241378c92ab'
down_revision = '12d3e8695f90'


def upgrade():
    """Add the type column to the builds table."""
    builds = sa.sql.table('builds', sa.sql.column('type', sa.Integer()))
    op.add_column('builds', sa.Column('type', sa.Integer(), nullable=True))
    # The type 1 is the RPM Build type.
    op.execute(builds.update().values({'type': 1}))
    op.alter_column('builds', 'type', nullable=False)


def downgrade():
    """Remove the type column from the builds table."""
    op.drop_column('builds', 'type')



# -*- coding: utf-8 -*-
# Copyright © 2013-2017 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""A set of API schemas to validate input and generate documentation."""
import re
import os

import colander

from bodhi.server import util
from bodhi.server.config import config
from bodhi.server.models import (
    ContentType,
    ReleaseState,
    UpdateRequest,
    UpdateSeverity,
    UpdateStatus,
    UpdateSuggestion,
    UpdateType,
)
from bodhi.server.validators import validate_csrf_token


CVE_REGEX = re.compile(r"CVE-[0-9]{4,4}-[0-9]{4,}")

# Retrieving list of templates from filesystem for `mail_template` validation in SaveReleaseSchema
template_directory = util.get_absolute_path(config.get('mail.templates_basepath'))
MAIL_TEMPLATES = [os.path.splitext(file)[0] for file in os.listdir(template_directory)]


class CSRFProtectedSchema(colander.MappingSchema):
    """A mixin class to validate csrf tokens."""

    csrf_token = colander.SchemaNode(
        colander.String(),
        name="csrf_token",
        validator=validate_csrf_token,
    )


class Bugs(colander.SequenceSchema):
    """A SequenceSchema to validate a list of Bug objects."""

    bug = colander.SchemaNode(colander.String(), missing=None)


class Builds(colander.SequenceSchema):
    """A SequenceSchema to validate a list of Build objects."""

    build = colander.SchemaNode(colander.String())


class CVE(colander.String):
    """A String schema to validate a CVE."""

    def deserialize(self, node, cstruct):
        """Parse a CVE out of a given API CVE parameter."""
        value = super(CVE, self).deserialize(node, cstruct)

        if CVE_REGEX.match(value) is None:
            raise colander.Invalid(node, '"%s" is not a valid CVE id' % value)

        return value


class CVEs(colander.SequenceSchema):
    """A SequenceSchema to validate a list of CVE objects."""

    cve = colander.SchemaNode(CVE())


class Packages(colander.SequenceSchema):
    """A SequenceSchema to validate a list of Package objects."""

    package = colander.SchemaNode(colander.String())


class Users(colander.SequenceSchema):
    """A SequenceSchema to validate a list of User objects."""

    user = colander.SchemaNode(colander.String())


class Releases(colander.SequenceSchema):
    """A SequenceSchema to validate a list of Release objects."""

    release = colander.SchemaNode(colander.String())


class ReleaseIds(colander.SequenceSchema):
    """A SequenceSchema to validate a list of Release ID objects."""

    release_id = colander.SchemaNode(colander.Integer())


class Groups(colander.SequenceSchema):
    """A SequenceSchema to validate a list of Group objects."""

    group = colander.SchemaNode(colander.String())


class Updates(colander.SequenceSchema):
    """A SequenceSchema to validate a list of Update objects."""

    update = colander.SchemaNode(colander.String())


class Tests(colander.SequenceSchema):
    """A SequenceSchema to validate a list of Test objects."""

    test = colander.SchemaNode(colander.String())


class BugFeedback(colander.MappingSchema):
    """A schema for BugFeedback to be provided via API parameters."""

    bug_id = colander.SchemaNode(colander.Integer())
    karma = colander.SchemaNode(
        colander.Integer(),
        validator=colander.Range(min=-1, max=1),
        missing=0,
    )


class BugFeedbacks(colander.SequenceSchema):
    """A SequenceSchema to validate a list of BugFeedback objects."""

    bug_feedback = BugFeedback()


class TestcaseFeedback(colander.MappingSchema):
    """A schema for TestcaseFeedback to be provided via API parameters."""

    testcase_name = colander.SchemaNode(colander.String())
    karma = colander.SchemaNode(
        colander.Integer(),
        validator=colander.Range(min=-1, max=1),
        missing=0,
    )


class TestcaseFeedbacks(colander.SequenceSchema):
    """A SequenceSchema to validate a list of TestcaseFeedback objects."""

    testcase_feedback = TestcaseFeedback()


class SaveCommentSchema(CSRFProtectedSchema, colander.MappingSchema):
    """An API schema for bodhi.server.services.comments.new_comment()."""

    def deserialize(self, cstruct):
        """Unflatten comment before parsing into Schema."""
        appstruct = SaveCommentSchema().unflatten(cstruct)
        return super(SaveCommentSchema, self).deserialize(appstruct)

    update = colander.SchemaNode(colander.String())
    text = colander.SchemaNode(
        colander.String(),
        missing='',
    )
    karma = colander.SchemaNode(
        colander.Integer(),
        validator=colander.Range(min=-1, max=1),
        missing=0,
    )
    karma_critpath = colander.SchemaNode(
        colander.Integer(),
        validator=colander.Range(min=-1, max=1),
        missing=0,
    )
    bug_feedback = BugFeedbacks(missing=[])
    testcase_feedback = TestcaseFeedbacks(missing=[])

    # Optional
    captcha_key = colander.SchemaNode(colander.String(), missing=None)
    captcha_value = colander.SchemaNode(colander.String(), missing=None)
    email = colander.SchemaNode(
        colander.String(),
        validator=colander.Email(),
        missing=None,
    )


class SaveUpdateSchema(CSRFProtectedSchema, colander.MappingSchema):
    """An API schema for bodhi.server.services.updates.new_update()."""

    builds = Builds(colander.Sequence(accept_scalar=True),
                    preparer=[util.splitter])

    bugs = Bugs(colander.Sequence(accept_scalar=True), missing=None, preparer=[util.splitter])

    close_bugs = colander.SchemaNode(
        colander.Boolean(),
        missing=True,
    )
    type = colander.SchemaNode(
        colander.String(),
        validator=colander.OneOf(list(UpdateType.values())),
    )
    request = colander.SchemaNode(
        colander.String(),
        validator=colander.OneOf(list(UpdateRequest.values())),
        missing='testing',
    )
    severity = colander.SchemaNode(
        colander.String(),
        validator=colander.OneOf(list(UpdateSeverity.values())),
        missing='unspecified',
    )
    notes = colander.SchemaNode(
        colander.String(),
        validator=colander.Length(min=2),
        missing_msg='A description is required for the update.'
    )
    autokarma = colander.SchemaNode(
        colander.Boolean(),
        missing=True,
    )
    stable_karma = colander.SchemaNode(
        colander.Integer(),
        validator=colander.Range(min=1),
        missing=3,
    )
    unstable_karma = colander.SchemaNode(
        colander.Integer(),
        validator=colander.Range(max=-1),
        missing=-3,
    )
    suggest = colander.SchemaNode(
        colander.String(),
        validator=colander.OneOf(list(UpdateSuggestion.values())),
        missing='unspecified',
    )
    edited = colander.SchemaNode(
        colander.String(),
        missing='',
    )
    requirements = colander.SchemaNode(
        colander.String(),
        missing=None,
    )
    require_bugs = colander.SchemaNode(
        colander.Boolean(),
        missing=True,
    )
    require_testcases = colander.SchemaNode(
        colander.Boolean(),
        missing=True,
    )


class Cosmetics(colander.MappingSchema):
    """A mixin class used by schemas to validate the ``display_user`` API parameter."""

    display_user = colander.SchemaNode(
        colander.Boolean(true_choices=('true', '1')),
        location="querystring",
        missing=True,
    )


class PaginatedSchema(colander.MappingSchema):
    """A mixin class used by schemas to provide pagination support for API endpoints."""

    chrome = colander.SchemaNode(
        colander.Boolean(true_choices=('true', '1')),
        location="querystring",
        missing=True,
    )

    page = colander.SchemaNode(
        colander.Integer(),
        validator=colander.Range(min=1),
        location="querystring",
        missing=1,
    )

    rows_per_page = colander.SchemaNode(
        colander.Integer(),
        validator=colander.Range(min=1, max=1000),
        location="querystring",
        missing=20,
    )


class SearchableSchema(colander.MappingSchema):
    """A mixin class used by schemas to provide search support for API endpoints."""

    like = colander.SchemaNode(
        colander.String(),
        location="querystring",
        missing=None,
        preparer=[lambda x: x.strip() if x else x],
    )

    search = colander.SchemaNode(
        colander.String(),
        location="querystring",
        missing=None,
        preparer=[lambda x: x.strip() if x else x],
    )


class ListReleaseSchema(PaginatedSchema):
    """
    An API schema for listing releases.

    This schema is used by bodhi.server.services.releases.query_releases_html() and
    bodhi.server.services.releases.query_releases_json().
    """

    ids = ReleaseIds(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    name = colander.SchemaNode(
        colander.String(),
        location="querystring",
        missing=None,
    )

    updates = Updates(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    packages = Packages(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    state = colander.SchemaNode(
        colander.String(),
        validator=colander.OneOf(list(ReleaseState.values())),
        missing=None,
    )

    exclude_archived = colander.SchemaNode(
        colander.Boolean(true_choices=('true', '1')),
        location="querystring",
        missing=None,
    )


class SaveReleaseSchema(CSRFProtectedSchema, colander.MappingSchema):
    """An API schema for bodhi.server.services.releases.save_release()."""

    name = colander.SchemaNode(
        colander.String(),
    )
    long_name = colander.SchemaNode(
        colander.String(),
    )
    version = colander.SchemaNode(
        colander.String(),
        missing=None,
    )
    branch = colander.SchemaNode(
        colander.String(),
    )
    id_prefix = colander.SchemaNode(
        colander.String(),
    )
    dist_tag = colander.SchemaNode(
        colander.String(),
    )
    stable_tag = colander.SchemaNode(
        colander.String(),
    )
    testing_tag = colander.SchemaNode(
        colander.String(),
    )
    candidate_tag = colander.SchemaNode(
        colander.String(),
    )
    pending_signing_tag = colander.SchemaNode(
        colander.String(),
        missing="",
    )
    pending_testing_tag = colander.SchemaNode(
        colander.String(),
        missing="",
    )
    pending_stable_tag = colander.SchemaNode(
        colander.String(),
        missing="",
    )
    override_tag = colander.SchemaNode(
        colander.String(),
    )
    state = colander.SchemaNode(
        colander.String(),
        validator=colander.OneOf(list(ReleaseState.values())),
        missing="disabled",
    )
    edited = colander.SchemaNode(
        colander.String(),
        missing=None,
    )
    mail_template = colander.SchemaNode(
        colander.String(),
        missing="fedora_errata_template",
        validator=colander.OneOf(MAIL_TEMPLATES)
    )
    composed_by_bodhi = colander.SchemaNode(
        colander.Boolean(true_choices=('true', '1')),
        missing=True,
    )


class ListStackSchema(PaginatedSchema, SearchableSchema):
    """An API schema for bodhi.server.services.stacks.query_stacks()."""

    name = colander.SchemaNode(
        colander.String(),
        location="querystring",
        missing=None,
    )

    packages = Packages(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )


class SaveStackSchema(CSRFProtectedSchema, colander.MappingSchema):
    """An API schema for bodhi.server.services.stacks.save_stack()."""

    name = colander.SchemaNode(
        colander.String(),
    )

    packages = Packages(
        colander.Sequence(accept_scalar=True),
        missing=None,
        preparer=[util.splitter],
    )

    description = colander.SchemaNode(
        colander.String(),
        missing=None,
    )

    requirements = colander.SchemaNode(
        colander.String(),
        missing=None,
    )


class ListUserSchema(PaginatedSchema, SearchableSchema):
    """An API schema for bodhi.server.services.user.query_users()."""

    name = colander.SchemaNode(
        colander.String(),
        location="querystring",
        missing=None,
    )

    groups = Groups(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    updates = Updates(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    packages = Packages(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )


class ListUpdateSchema(PaginatedSchema, SearchableSchema, Cosmetics):
    """An API schema for bodhi.server.services.updates.query_updates()."""

    alias = Builds(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    approved_since = colander.SchemaNode(
        colander.DateTime(),
        location="querystring",
        missing=None,
    )

    approved_before = colander.SchemaNode(
        colander.DateTime(),
        location="querystring",
        missing=None,
    )

    bugs = Bugs(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    builds = Builds(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    critpath = colander.SchemaNode(
        colander.Boolean(true_choices=('true', '1')),
        location="querystring",
        missing=None,
    )

    cves = CVEs(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    locked = colander.SchemaNode(
        colander.Boolean(true_choices=('true', '1')),
        location="querystring",
        missing=None,
    )

    modified_since = colander.SchemaNode(
        colander.DateTime(),
        location="querystring",
        missing=None,
    )

    modified_before = colander.SchemaNode(
        colander.DateTime(),
        location="querystring",
        missing=None,
    )

    active_releases = Releases(
        colander.Boolean(true_choices=('true', '1')),
        location="querystring",
        missing=False,
    )

    packages = Packages(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    pushed = colander.SchemaNode(
        colander.Boolean(true_choices=('true', '1')),
        location="querystring",
        missing=None,
    )

    pushed_since = colander.SchemaNode(
        colander.DateTime(),
        location="querystring",
        missing=None,
    )

    pushed_before = colander.SchemaNode(
        colander.DateTime(),
        location="querystring",
        missing=None,
    )

    releases = Releases(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    # This singular version of the plural "releases" is purely for bodhi1
    # backwards compat (mostly for RSS feeds) - threebean
    release = colander.SchemaNode(
        colander.String(),
        location="querystring",
        missing=None,
    )

    request = colander.SchemaNode(
        colander.String(),
        location="querystring",
        missing=None,
        validator=colander.OneOf(list(UpdateRequest.values())),
    )

    severity = colander.SchemaNode(
        colander.String(),
        location="querystring",
        missing=None,
        validator=colander.OneOf(list(UpdateSeverity.values())),
    )

    status = colander.SchemaNode(
        colander.String(),
        location="querystring",
        missing=None,
        validator=colander.OneOf(list(UpdateStatus.values())),
    )

    submitted_since = colander.SchemaNode(
        colander.DateTime(),
        location="querystring",
        missing=None,
    )

    submitted_before = colander.SchemaNode(
        colander.DateTime(),
        location="querystring",
        missing=None,
    )

    suggest = colander.SchemaNode(
        colander.String(),
        location="querystring",
        missing=None,
        validator=colander.OneOf(list(UpdateSuggestion.values())),
    )

    type = colander.SchemaNode(
        colander.String(),
        location="querystring",
        missing=None,
        validator=colander.OneOf(list(UpdateType.values())),
    )

    content_type = colander.SchemaNode(
        colander.String(),
        location="querystring",
        missing=None,
        validator=colander.OneOf(list(ContentType.values())),
    )

    user = Users(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    updateid = Builds(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )


class ListPackageSchema(PaginatedSchema, SearchableSchema):
    """An API schema for bodhi.server.services.packages.query_packages()."""

    name = colander.SchemaNode(
        colander.String(),
        location="querystring",
        missing=None,
    )


class ListBuildSchema(PaginatedSchema):
    """An API schema for bodhi.server.services.builds.query_builds()."""

    nvr = colander.SchemaNode(
        colander.String(),
        location="querystring",
        missing=None,
    )

    packages = Packages(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    releases = Releases(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    updates = Updates(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )


class UpdateRequestSchema(CSRFProtectedSchema, colander.MappingSchema):
    """An API schema for bodhi.server.services.updates.set_request()."""

    request = colander.SchemaNode(
        colander.String(),
        validator=colander.OneOf(list(UpdateRequest.values())),
    )


class ListCommentSchema(PaginatedSchema, SearchableSchema):
    """An API schema for bodhi.server.services.comments.query_comments()."""

    updates = Updates(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    packages = Packages(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    user = Users(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    update_owner = Users(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    ignore_user = Users(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    anonymous = colander.SchemaNode(
        colander.Boolean(true_choices=('true', '1')),
        location="querystring",
        missing=None,
    )

    since = colander.SchemaNode(
        colander.DateTime(),
        location="querystring",
        missing=None,
    )


class ListOverrideSchema(PaginatedSchema, SearchableSchema, Cosmetics):
    """An API schema for bodhi.server.services.overrides.query_overrides()."""

    builds = Builds(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    expired = colander.SchemaNode(
        colander.Boolean(true_choices=('true', '1')),
        location="querystring",
        missing=None,
    )

    packages = Packages(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    releases = Releases(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )

    user = Users(
        colander.Sequence(accept_scalar=True),
        location="querystring",
        missing=None,
        preparer=[util.splitter],
    )


class SaveOverrideSchema(CSRFProtectedSchema, colander.MappingSchema):
    """An API schema for bodhi.server.services.overrides.save_override()."""

    nvr = colander.SchemaNode(
        colander.String(),
    )

    notes = colander.SchemaNode(
        colander.String(),
        validator=colander.Length(min=2),
    )

    expiration_date = colander.SchemaNode(
        colander.DateTime(default_tzinfo=None),
    )

    expired = colander.SchemaNode(
        colander.Boolean(),
        missing=False,
    )

    edited = colander.SchemaNode(
        colander.String(),
        missing=None,
    )


class WaiveTestResultsSchema(CSRFProtectedSchema, colander.MappingSchema):
    """An API schema for bodhi.server.services.updates.waive_test_results()."""

    comment = colander.SchemaNode(
        colander.String(),
        missing=None,
    )
    tests = Tests(colander.Sequence(accept_scalar=True), missing=None, preparer=[util.splitter])


class GetTestResultsSchema(CSRFProtectedSchema, colander.MappingSchema):
    """An API schema for bodhi.server.services.updates.get_test_results()."""

    alias = Builds(
        colander.Sequence(accept_scalar=True),
        missing=None,
    )



import logging
import os
import subprocess

from fedora.client import BodhiClient, BodhiClientException
from fedora.client.bodhi import Bodhi2Client


log = logging.getLogger('fedora.client.bodhi')

build = u'qt-creator-3.4.1-3.fc23'
username = os.getenv('USER')
p = subprocess.Popen(['/usr/bin/pass', 'fedora'], stdout=subprocess.PIPE)
out, err = p.communicate()
password = out.strip()

print('Logging into bodhi2')
bodhi = BodhiClient(staging=True, username=username, password=password)

print('Logged in! Creating new update...')

try:
    result = bodhi.save(
        builds=build, type='bugfix', notes='The quick brown fox jumped over the lazy dog',
        bugs='1234106,1234107')
    print(result)
    assert len(result.bugs) == 2, result.bugs
except BodhiClientException as e:
    print(e)

print('Editing update')
try:
    result = bodhi.save(
        builds=build,
        edited=build,
        type='enhancement',
        notes='The quick brown fox jumped over the lazy dog',
        bugs='1234106',
        require_bugs=True)
    assert len(result.bugs) == 1, result.bugs
    assert result.require_bugs, result
    for caveat in result.caveats:
        print(caveat.description)
except BodhiClientException as e:
    print(e)


print('Querying update')
result = bodhi.query(builds='qt-creator-3.4.1-3.fc23')
updates = result['updates']
update = updates[0]
assert len(updates) == 1, len(updates)
assert result['total'] == 1, result
alias = update['alias']
print(alias)

print(bodhi.update_str(update))
print('')
print(bodhi.update_str(update, minimal=True))

print('Call /latest_builds')
result = bodhi.latest_builds('kernel')
print(result)

print('Querying all releases')
result = bodhi.get_releases()
print(result)

print('Looking for candidate builds')
print(bodhi.candidates())

print('Looking for local builds in updates-testing')
for update in bodhi.testable():
    print(bodhi.update_str(update))

print('Querying by release')
result = bodhi.query(release='F23')
print(result)
updates = result['updates']
print(updates)
assert len(updates) == 1, len(updates)
assert result['total'] == 1, result
assert updates[0].alias == alias
print('%d updates returned' % len(updates))

print('Querying by release and package')
result = bodhi.query(package='qt-creator', release='F23')
updates = result['updates']
assert len(updates) == 1, len(updates)
assert result['total'] == 1, result
assert updates[0]['alias'] == alias
print('%d updates returned' % len(updates))

print('Querying by release and package')
result = bodhi.query(package='qt-creator', release='F23')
updates = result.updates
assert len(updates) == 1, len(updates)
assert result['total'] == 1, result
assert updates[0]['alias'] == alias
print('%d updates returned' % len(updates))

print('Querying by release and package and status and limit')
result = bodhi.query(type='security', limit=2)
updates = result.updates
assert len(updates) == 2, len(updates)
assert updates[0]['type'] == 'security'
print('%d updates returned' % len(updates))

print('Querying by release with limit (blockerbugs)')
result = bodhi.query(limit=100, release='F23')
updates = result['updates']
assert len(updates) == 1, len(updates)
assert updates[0]['type'] == 'bugfix', updates[0].type
print('%d updates returned' % len(updates))

print('Requesting stable')
try:
    result = bodhi.request(update=alias, request='stable')
    assert False, "Client didn't throw exception"
except BodhiClientException as e:
    print(e)

print('Requesting testing')
result = bodhi.request(update=alias, request='testing')
assert result['update']['request'] == 'testing'

print('Revoking request')
result = bodhi.request(update=alias, request='revoke')
assert result['update']['request'] is None

print('Adding comment')
result = bodhi.comment(update=alias, comment='yay', karma=1)
print(result)
comment = result['comment']
assert comment['author'] == username
assert comment['text'] == u'yay'
assert comment['update']['title'] == build

print('Querying multiple pages of updates')
query = dict(limit=100, release='F22', critpath=True)
result = bodhi.query(**query)
updates = result.updates
total = result.total
page = result.page
pages = result.pages
print('%r pages' % result.pages)
print('%r page' % result.page)
print('%r total' % result.total)
while result.page < result.pages:
    print('Querying page %d out of %d' % (result.page + 1, pages))
    result = bodhi.query(page=result.page + 1, **query)
    updates.extend(result.updates)

print('Fetched %d updates total' % len(updates))
assert len(updates) == total, len(updates)

print('Querying for my updates')
result = bodhi.query(mine=True)
print(result)
assert result.updates[0].user.name == username, result.updates[0].user.name


print('Testing the Bodhi2Client(staging=True) directly')
bodhi = Bodhi2Client(staging=True, username=username, password=password)
try:
    result = bodhi.save(
        builds=build,
        type='bugfix',
        notes='The quick brown fox jumped over the lazy dog',
    )
    print(result)
except BodhiClientException as e:
    print(e)

result = bodhi.query(builds='qt-creator-3.4.1-3.fc23')
updates = result['updates']
update = updates[0]
assert len(updates) == 1, len(updates)



# Copyright (c) 2018 Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Drop the uniqueness constraint on releases.branch.

Revision ID: adee0d22d09f
Revises: 2616c86d8ac6
Create Date: 2018-03-13 20:55:03.345503
"""
from alembic import op


# revision identifiers, used by Alembic.
revision = 'adee0d22d09f'
down_revision = '2616c86d8ac6'


def upgrade():
    """Drop the uniqueness constraint on releases.branch."""
    op.drop_constraint(u'releases_branch_key', 'releases', type_='unique')


def downgrade():
    """Recreate the uniqueness constraint on releases.branch."""
    op.create_unique_constraint(u'releases_branch_key', 'releases', ['branch'])



# -*- coding: utf-8 -*-
# Copyright © 2015-2017 Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Define a service endpoint for searching for packages."""
import math

from cornice import Service
from cornice.validators import colander_querystring_validator
from sqlalchemy import func, distinct
from sqlalchemy.sql.expression import case

from bodhi.server.models import Package
import bodhi.server.schemas
import bodhi.server.security
import bodhi.server.services.errors


packages = Service(name='packages', path='/packages/',
                   description='PkgDB packages',
                   cors_origins=bodhi.server.security.cors_origins_ro)


@packages.get(
    schema=bodhi.server.schemas.ListPackageSchema, renderer='json',
    error_handler=bodhi.server.services.errors.json_handler,
    validators=(colander_querystring_validator,))
def query_packages(request):
    """
    Search for packages via query string parameters.

    The following query string parameters may be used to limit the packages returned by the service:
        name: The name of the Packages you wish to retrieve.
        like: Search for Packages with names like the given string.
        search: Search for Packages with names like the given string, with case insensitivity.
        page: Retrieve the specified page of search results.
        rows_per_page: Specify how many rows per page are desired.

    Args:
        request (pyramid.request): The current web request.
    Returns:
        dict: A dictionary with the following key value mappings:
            packages: An iterable of packages that match the search criteria.
            page: The current page of results.
            pages: The number of pages of results.
            rows_per_page: The number of results per page.
            total: The total number of packages that match the search criteria.
    """
    db = request.db
    data = request.validated
    query = db.query(Package)

    name = data.get('name')
    if name is not None:
        query = query.filter(Package.name == name)

    like = data.get('like')
    if like is not None:
        query = query.filter(Package.name.like('%%%s%%' % like))

    search = data.get('search')
    if search is not None:
        query = query.filter(Package.name.ilike('%%%s%%' % search))
        query = query.order_by(case([(Package.name == search, Package.name)]))

    # We can't use ``query.count()`` here because it is naive with respect to
    # all the joins that we're doing above.
    count_query = query.with_labels().statement\
        .with_only_columns([func.count(distinct(Package.name))])\
        .order_by(None)
    total = db.execute(count_query).scalar()

    page = data.get('page')
    rows_per_page = data.get('rows_per_page')
    pages = int(math.ceil(total / float(rows_per_page)))
    query = query.offset(rows_per_page * (page - 1)).limit(rows_per_page)

    return dict(
        packages=query.all(),
        page=page,
        pages=pages,
        rows_per_page=rows_per_page,
        total=total,
    )



# -*- coding: utf-8 -*-
# Copyright © 2016-2018 Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Cleans up old mashes that are left over in mash_dir."""
import collections

import click
import os
import shutil

from bodhi.server import config


# How many of the newest mash dirs to keep during cleanup
NUM_TO_KEEP = 10


@click.command()
@click.version_option(message='%(version)s')
def clean_up():
    """Delete any repo mashes that are older than the newest 10 from each repo series."""
    remove_old_composes()


# Helper function used in auto clean composes (masher.py)
def remove_old_composes():
    """Delete any repo mashes that are older than the newest 10 from each repo series."""
    mash_dir = config.config['mash_dir']

    # This data structure will map the beginning of a group of dirs for the same repo to a list of
    # the dirs that start off the same way.
    pattern_matched_dirs = collections.defaultdict(list)

    for directory in [d for d in os.listdir(mash_dir) if os.path.isdir(os.path.join(mash_dir, d))]:
        # If this directory ends with a float, it is a candidate for potential deletion
        try:
            split_dir = directory.split('-')
            float(split_dir[-1])
        except ValueError:
            # This directory didn't end in a float, so let's just move on to the next one.
            continue

        pattern = directory.replace(split_dir[-1], '')
        pattern_matched_dirs[pattern].append(directory)

    dirs_to_delete = []

    for dirs in pattern_matched_dirs.values():
        if len(dirs) > NUM_TO_KEEP:
            dirs_to_delete.extend(sorted(dirs, reverse=True)[NUM_TO_KEEP:])

    if dirs_to_delete:
        print('Deleting the following directories:')
        for d in dirs_to_delete:
            d = os.path.join(mash_dir, d)
            print(d)
            shutil.rmtree(d)
